# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch

from typing import List, Dict, Any

import embodichain.lab.gym.envs.managers.randomization as rand
import embodichain.lab.gym.envs.managers.events as events
import embodichain.lab.gym.envs.managers.observations as obs

from embodichain.lab.gym.envs.managers import (
    EventCfg,
    SceneEntityCfg,
    ObservationCfg,
)
from embodichain.lab.gym.envs import EmbodiedEnv, EmbodiedEnvCfg
from embodichain.lab.gym.utils.registration import register_env
from embodichain.lab.sim.robots import DexforceW1Cfg
from embodichain.lab.sim.sensors import StereoCameraCfg, SensorCfg
from embodichain.lab.sim.shapes import MeshCfg
from embodichain.lab.sim.cfg import (
    LightCfg,
    ArticulationCfg,
    RobotCfg,
    RigidObjectCfg,
    RigidBodyAttributesCfg,
)
from embodichain.data import get_data_path
from embodichain.utils import configclass


@configclass
class ExampleEventCfg:

    replace_obj: EventCfg = EventCfg(
        func=events.replace_assets_from_group,
        mode="reset",
        params={
            "entity_cfg": SceneEntityCfg(
                uid="fork",
            ),
            "folder_path": get_data_path("TableWare/tableware/fork/"),
        },
    )

    randomize_fork_mass: EventCfg = EventCfg(
        func=rand.randomize_rigid_object_mass,
        mode="reset",
        params={
            "entity_cfg": SceneEntityCfg(
                uid="fork",
            ),
            "mass_range": (0.1, 2.0),
        },
    )

    randomize_light: EventCfg = EventCfg(
        func=rand.randomize_light,
        mode="interval",
        interval_step=5,
        params={
            "entity_cfg": SceneEntityCfg(
                uid="point",
            ),
            "position_range": [[-0.5, -0.5, 2], [0.5, 0.5, 2]],
            "color_range": [[0.6, 0.6, 0.6], [1, 1, 1]],
            "intensity_range": [50.0, 100.0],
        },
    )

    randomize_table_mat: EventCfg = EventCfg(
        func=rand.randomize_visual_material,
        mode="interval",
        interval_step=10,
        params={
            "entity_cfg": SceneEntityCfg(
                uid="table",
            ),
            "random_texture_prob": 0.5,
            "texture_path": get_data_path("CocoBackground/coco"),
            "base_color_range": [[0.2, 0.2, 0.2], [1.0, 1.0, 1.0]],
        },
    )


@configclass
class ObsCfg:

    obj_pose: ObservationCfg = ObservationCfg(
        func=obs.get_rigid_object_pose,
        mode="add",
        name="fork_pose",
        params={"entity_cfg": SceneEntityCfg(uid="fork")},
    )


@configclass
class ExampleCfg(EmbodiedEnvCfg):

    # Define the robot configuration using DexforceW1Cfg
    robot: RobotCfg = DexforceW1Cfg.from_dict(
        {
            "uid": "dexforce_w1",
            "version": "v021",
            "arm_kind": "anthropomorphic",
            "init_pos": [0.0, 0, 0.0],
        }
    )

    # Define the sensor configuration using StereoCameraCfg
    sensor: List[SensorCfg] = [
        StereoCameraCfg(
            uid="eye_in_head",
            width=960,
            height=540,
            enable_mask=True,
            enable_depth=True,
            left_to_right_pos=(0.06, 0, 0),
            intrinsics=(450, 450, 480, 270),
            intrinsics_right=(450, 450, 480, 270),
            extrinsics=StereoCameraCfg.ExtrinsicsCfg(
                parent="eyes",
            ),
        )
    ]

    light: EmbodiedEnvCfg.EnvLightCfg = EmbodiedEnvCfg.EnvLightCfg(
        direct=[
            LightCfg(
                uid="point",
                light_type="point",
                color=(1.0, 1.0, 1.0),
                intensity=50.0,
                init_pos=(0, 0, 2),
            )
        ]
    )

    background: List[RigidObjectCfg] = [
        RigidObjectCfg(
            uid="table",
            shape=MeshCfg(
                fpath=get_data_path("CircleTableSimple/circle_table_simple.ply"),
                compute_uv=True,
            ),
            attrs=RigidBodyAttributesCfg(
                mass=10.0,
                static_friction=0.95,
                dynamic_friction=0.85,
                restitution=0.01,
            ),
            body_type="kinematic",
            init_pos=(0.80, 0, 0.8),
            init_rot=(0, 90, 0),
        ),
    ]

    rigid_object: List[RigidObjectCfg] = [
        RigidObjectCfg(
            uid="fork",
            shape=MeshCfg(
                fpath=get_data_path("TableWare/tableware/fork/standard_fork_scale.ply"),
            ),
            body_scale=(0.75, 0.75, 1.0),
            init_pos=(0.8, 0, 1.0),
        ),
    ]

    articulation_cfg: List[ArticulationCfg] = [
        ArticulationCfg(
            uid="drawer",
            fpath="SlidingBoxDrawer/SlidingBoxDrawer.urdf",
            init_pos=(0.5, 0.0, 0.85),
        )
    ]

    events = ExampleEventCfg()

    observations = ObsCfg()


@register_env("ModularEnv-v1", max_episode_steps=100, override=True)
class ModularEnv(EmbodiedEnv):
    """
    An example of a modular environment that inherits from EmbodiedEnv
    and uses custom event and observation managers.
    """

    def __init__(self, cfg: EmbodiedEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)


if __name__ == "__main__":
    import gymnasium as gym
    import argparse

    from embodichain.lab.sim import SimulationManagerCfg

    parser = argparse.ArgumentParser()
    parser.add_argument("--enable_rt", action="store_true", help="Enable ray tracing")
    args = parser.parse_args()

    env_cfg = ExampleCfg(sim_cfg=SimulationManagerCfg(enable_rt=args.enable_rt))

    # Create the Gym environment
    env = gym.make("ModularEnv-v1", cfg=env_cfg)

    while True:
        obs, info = env.reset()

        for i in range(100):
            action = torch.zeros(env.action_space.shape, dtype=torch.float32)
            obs, reward, done, truncated, info = env.step(action)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import os
import time
import numpy as np
import torch
from IPython import embed

from embodichain.data import get_data_path
from embodichain.lab.sim.cfg import RobotCfg
from embodichain.lab.sim.objects import Robot
from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import MarkerCfg


def main(visualize: bool = True):
    np.set_printoptions(precision=5, suppress=True)
    torch.set_printoptions(precision=5, sci_mode=False)

    # Set up simulation with specified device (CPU or CUDA)
    sim_device = "cpu"
    num_envs = 9  # Number of parallel arenas/environments
    config = SimulationManagerCfg(
        headless=False, sim_device=sim_device, arena_space=1.5, num_envs=num_envs
    )
    sim = SimulationManager(config)
    sim.set_manual_update(False)

    # Load robot URDF file
    urdf = get_data_path("Rokae/SR5/SR5.urdf")
    assert os.path.isfile(urdf)

    # Robot configuration
    cfg_dict = {
        "fpath": urdf,
        "control_parts": {
            "main_arm": [
                "joint1",
                "joint2",
                "joint3",
                "joint4",
                "joint5",
                "joint6",
            ],
        },
        "solver_cfg": {
            "main_arm": {
                "class_type": "DifferentialSolver",
                "end_link_name": "ee_link",
                "root_link_name": "base_link",
            },
        },
    }

    robot: Robot = sim.add_robot(cfg=RobotCfg.from_dict(cfg_dict))

    # Prepare initial joint positions for all environments
    rad = torch.deg2rad(torch.tensor(45.0))
    arm_name = "main_arm"
    fk_qpos = torch.full((num_envs, 6), rad, dtype=torch.float32, device="cpu")
    # All envs start with the same qpos (can be randomized)
    qpos = torch.from_numpy(np.array([0.0, 0.0, np.pi / 2, 0.0, np.pi / 2, 0.0])).to(
        fk_qpos.device
    )
    qpos = qpos.unsqueeze(0).repeat(num_envs, 1)
    robot.set_qpos(qpos=qpos, joint_ids=robot.get_joint_ids(arm_name))

    time.sleep(3.0)
    fk_xpos = robot.compute_fk(
        qpos=qpos, name=arm_name, to_matrix=True
    )  # (num_envs, 4, 4)

    # Prepare batch start and end poses for all envs
    start_pose = fk_xpos.clone()  # (num_envs, 4, 4)
    end_pose = fk_xpos.clone()
    move_vecs = torch.tensor(
        [
            [0.3, 0.0, 0.0],
            [0.2, -0.2, 0.0],
            [0.0, 0.0, 0.2],
            [0.2, 0.0, 0.2],
            [-0.3, 0.0, 0.0],
            [-0.2, 0.2, 0.0],
            [0.0, 0.0, -0.2],
            [-0.2, 0.0, -0.2],
            [0.1, 0.1, -0.1],
        ],
        dtype=end_pose.dtype,
        device=end_pose.device,
    )
    end_pose[
        :, :3, 3
    ] += move_vecs  # Move each env's end-effector in a different direction

    num_steps = 100
    # Interpolate poses for each env
    interpolated_poses = torch.stack(
        [torch.lerp(start_pose, end_pose, t) for t in np.linspace(0, 1, num_steps)],
        dim=1,
    )  # (num_envs, num_steps, 4, 4)

    # Initial joint positions for all envs
    ik_qpos = qpos.clone()

    ik_qpos_results = []
    ik_success_flags = []

    ik_compute_begin = time.time()
    # Batch IK solving for each step
    for step in range(num_steps):
        poses = interpolated_poses[:, step, :, :]  # (num_envs, 4, 4)
        if poses.shape[0] != num_envs:
            poses = poses.expand(num_envs, *poses.shape[1:])
        if ik_qpos.shape[0] != num_envs:
            ik_qpos = ik_qpos.expand(num_envs, *ik_qpos.shape[1:])
        assert (
            poses.shape[0] == num_envs
        ), f"poses batch mismatch: {poses.shape[0]} vs {num_envs}"
        assert (
            ik_qpos.shape[0] == num_envs
        ), f"ik_qpos batch mismatch: {ik_qpos.shape[0]} vs {num_envs}"

        # Parallel batch IK solving
        res, ik_qpos_new = robot.compute_ik(
            pose=poses, joint_seed=ik_qpos, name=arm_name
        )
        ik_qpos_results.append(ik_qpos_new.clone())
        ik_success_flags.append(res)
        ik_qpos = ik_qpos_new  # Update joint seed
    ik_compute_end = time.time()
    print(
        f"IK compute time for {num_steps} steps and {num_envs} envs: {ik_compute_end - ik_compute_begin:.4f} seconds"
    )

    # Collect visualization data for all steps and environments
    if visualize:
        draw_data = [[] for _ in range(num_envs)]
    for env_id in range(num_envs):
        for step in range(num_steps):
            ik_qpos_new = ik_qpos_results[step]
            ik_xpos = robot.compute_fk(qpos=ik_qpos_new, name=arm_name, to_matrix=True)
            local_pose = robot._entities[env_id].get_world_pose()
            if visualize:
                fk_axis = local_pose @ end_pose[env_id].cpu().numpy()
                ik_axis = local_pose @ ik_xpos[env_id].cpu().numpy()
                local_axis = local_pose @ ik_xpos[env_id].cpu().numpy()

                draw_data[env_id].append(
                    {
                        "step": step,
                        "fk_axis": fk_axis,
                        "ik_axis": ik_axis,
                        "local_axis": local_axis,
                    }
                )

    if visualize:
        # Batch draw all steps' data for each environment
        for env_id in range(num_envs):
            # Only draw fk_axis and ik_axis once per env (first step)
            fk_axis = draw_data[env_id][0]["fk_axis"]
            ik_axis = draw_data[env_id][0]["ik_axis"]

            sim.draw_marker(
                cfg=MarkerCfg(
                    name=f"fk_axis_env{env_id}",
                    marker_type="axis",
                    axis_xpos=fk_axis,
                    axis_size=0.002,
                    axis_len=0.005,
                    arena_index=env_id,
                )
            )

            sim.draw_marker(
                cfg=MarkerCfg(
                    name=f"ik_axis_env{env_id}",
                    marker_type="axis",
                    axis_xpos=ik_axis,
                    axis_size=0.002,
                    axis_len=0.005,
                    arena_index=env_id,
                )
            )

            # Draw the whole local_axis trajectory as a single call
            local_axes = np.stack(
                [item["local_axis"] for item in draw_data[env_id]], axis=0
            )  # (num_steps, 4, 4) or (num_steps, 3)

            sim.draw_marker(
                cfg=MarkerCfg(
                    name=f"local_axis_env{env_id}_trajectory",
                    marker_type="axis",
                    axis_xpos=local_axes,
                    axis_size=0.002,
                    axis_len=0.005,
                    arena_index=env_id,
                )
            )

    # Optionally, set qpos for each step (replay or animate)
    for step in range(num_steps):
        ik_qpos_new = ik_qpos_results[step]
        res = ik_success_flags[step]
        # Only set qpos for successful IK results
        if isinstance(res, (list, np.ndarray, torch.Tensor)):
            for env_id, success in enumerate(res):
                if success:
                    q = (
                        ik_qpos_new[env_id]
                        if ik_qpos_new.dim() == 3
                        else ik_qpos_new[env_id]
                    )
                    robot.set_qpos(
                        qpos=q,
                        joint_ids=robot.get_joint_ids(arm_name),
                        env_ids=[env_id],
                    )
        else:
            # fallback: set all
            if ik_qpos_new.dim() == 3:
                robot.set_qpos(
                    qpos=ik_qpos_new[:, 0, :], joint_ids=robot.get_joint_ids(arm_name)
                )
            else:
                robot.set_qpos(
                    qpos=ik_qpos_new, joint_ids=robot.get_joint_ids(arm_name)
                )
        time.sleep(0.005)

    embed(header="Test DifferentialSolver example. Press Ctrl+D to exit.")


if __name__ == "__main__":
    main(visualize=True)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
This script demonstrates how to create a simulation scene using SimulationManager.
It supports loading kitchen/factory/office scenes via EmbodiChainDataset.
"""

import argparse
import time
from pathlib import Path
import math
import embodichain.utils.logger as logger
from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import RigidBodyAttributesCfg, LightCfg, RobotCfg, URDFCfg
from embodichain.lab.sim.shapes import MeshCfg
from embodichain.lab.sim.objects import RigidObject, RigidObjectCfg, Robot
from embodichain.data.assets.scene_assets import SceneData
from embodichain.data.constants import EMBODICHAIN_DEFAULT_DATA_ROOT


def resolve_asset_path(scene_name: str) -> str:
    """
    Resolve the asset path for a given scene (.glb/.gltf),
    downloading if needed using EmbodiChainData.
    """

    current_dir = Path(__file__).parent
    local_glb = current_dir / f"{scene_name}.glb"
    local_gltf = current_dir / f"{scene_name}.gltf"
    if local_glb.exists():
        logger.log_info(f"Using local asset: {local_glb}")
        return str(local_glb)
    if local_gltf.exists():
        logger.log_info(f"Using local asset: {local_gltf}")
        return str(local_gltf)

    scene_data = SceneData()

    extracted_dir = Path(EMBODICHAIN_DEFAULT_DATA_ROOT) / "extract" / "SceneData"
    glb_path = extracted_dir / f"{scene_name}.glb"
    gltf_path = extracted_dir / f"{scene_name}.gltf"

    if glb_path.exists():
        logger.log_info(f"Using downloaded asset: {glb_path}")
        return str(glb_path)
    if gltf_path.exists():
        logger.log_info(f"Using downloaded asset: {gltf_path}")
        return str(gltf_path)

    raise FileNotFoundError(
        f"No .glb or .gltf found in extracted folder: {extracted_dir}"
    )


def run_simulation(sim: SimulationManager):
    """Run the simulation loop."""
    if sim.is_use_gpu_physics:
        sim.init_gpu_physics()

    try:
        while True:
            time.sleep(0.01)
    except KeyboardInterrupt:
        logger.log_info("\n Stopping simulation...")
    finally:
        sim.destroy()
        logger.log_info("Simulation terminated successfully.")


def main():
    parser = argparse.ArgumentParser(
        description="Create a simulation scene with SimulationManager"
    )
    parser.add_argument(
        "--scene",
        type=str,
        default="kitchen",
        choices=["kitchen", "factory", "office", "local"],
        help="Choose which scene to load",
    )
    parser.add_argument(
        "--num_envs", type=int, default=1, help="Number of parallel environments"
    )
    parser.add_argument(
        "--device", type=str, default="cpu", help="Simulation device (cuda or cpu)"
    )
    parser.add_argument(
        "--disable_rt",
        action="store_true",
        default=False,
        help="Disable ray tracing for better visuals",
    )
    args = parser.parse_args()

    logger.log_info(f"Initializing scene '{args.scene}'")

    logger.log_info(f"Resolving and downloading scene '{args.scene}' if needed...")
    try:
        asset_path = resolve_asset_path(args.scene)
        logger.log_info(f"Scene asset ready at: {asset_path}")
    except Exception as e:
        print(f"Failed to download or resolve scene asset: {e}")
        return

    sim_cfg = SimulationManagerCfg(
        width=1920,
        height=1080,
        headless=True,
        physics_dt=1.0 / 100.0,
        sim_device=args.device,
        enable_rt=not args.disable_rt,
        num_envs=args.num_envs,
        arena_space=10.0,
    )
    sim = SimulationManager(sim_cfg)

    num_lights = 8
    radius = 5
    height = 8
    intensity = 200
    lights = []

    for i in range(num_lights):
        angle = 2 * math.pi * i / num_lights
        x = radius * math.cos(angle)
        y = radius * math.sin(angle)
        z = height
        uid = f"l{i+1}"
        cfg = LightCfg(uid=uid, intensity=intensity, radius=600, init_pos=[x, y, z])
        lights.append(sim.add_light(cfg))

    physics_attrs = RigidBodyAttributesCfg(
        mass=10,
        dynamic_friction=0.5,
        static_friction=0.5,
        restitution=0.1,
    )

    try:
        logger.log_info(f"Loading scene asset into simulation: {asset_path}")
        scene_obj: RigidObject = sim.add_rigid_object(
            cfg=RigidObjectCfg(
                uid=args.scene,
                shape=MeshCfg(fpath=asset_path),
                body_type="static",
                init_pos=[0.0, 0.0, 0.1],
                init_rot=[90, 180, 0],
                attrs=physics_attrs,
            )
        )
        if args.scene == "factory":
            from embodichain.lab.sim.robots.dexforce_w1.cfg import DexforceW1Cfg

            w1_robot: Robot = sim.add_robot(
                cfg=DexforceW1Cfg.from_dict(
                    {
                        "uid": "dexforce_w1",
                        "version": "v021",
                        "arm_kind": "anthropomorphic",
                        "init_pos": [-1, -0.5, 0],
                        "init_rot": [0, 0, 90],
                    }
                ),
            )

    except Exception as e:
        logger.log_info(f"Failed to load scene asset: {e}")
        return

    logger.log_info(f"Scene '{args.scene}' setup complete!")
    logger.log_info(f"Running simulation with {args.num_envs} environment(s)")
    logger.log_info("Press Ctrl+C to stop the simulation")

    sim.open_window()

    run_simulation(sim)


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------


import os
import open3d as o3d
from embodichain.data.dataset import EmbodiChainDataset
from embodichain.data.constants import (
    EMBODICHAIN_DOWNLOAD_PREFIX,
    EMBODICHAIN_DEFAULT_DATA_ROOT,
)


robot_assets = "robot_assets"


class CobotMagicArm(EmbodiChainDataset):
    """Dataset class for the Cobot Magic Arm robot.

    Reference:
        https://global.agilex.ai/products/cobot-magic

    Directory structure:
        CobotMagicArm/
            CobotMagicNoGripper.urdf
            CobotMagicWithGripperV70.urdf
            CobotMagicWithGripperV70NewUV.urdf
            CobotMagicWithGripperV70NoMaterial.urdf
            CobotMagicWithGripperV100.urdf
            CobotMagicWithGripperV100NewUV.urdf
            CobotMagicWithGripperV100NoMaterial.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import CobotMagicArm
        >>> dataset = CobotMagicArm()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("CobotMagicArm/CobotMagicWithGripperV100.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(
                EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "CobotMagicArmV2.zip"
            ),
            "14af3e84b74193680899a59fc74e8337",
        )
        prefix = "CobotMagicArm"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class RidgeBack(EmbodiChainDataset):
    """Dataset class for the RidgeBack wheeled robot.

    Reference:
        https://clearpathrobotics.com/ridgeback-indoor-robot-platform/

    Directory structure:
        RidgeBack/
            RidgeBack.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import RidgeBack
        >>> dataset = RidgeBack()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("RidgeBack/RidgeBack.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "RidgeBack.zip"),
            "f03e1a6f4c781ad8957a88bdb010e9b6",
        )
        prefix = "RidgeBack"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class UnitreeH1(EmbodiChainDataset):
    """Dataset class for the Unitree H1 robot.

    Reference:
        https://www.unitree.com/h1/

    Directory structure:
        UnitreeH1/
            UnitreeH1.urdf
            UnitreeH1WithWrist.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import UnitreeH1
        >>> dataset = UnitreeH1()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("UnitreeH1/UnitreeH1.urdf"))
        >>> print(get_data_path("UnitreeH1/UnitreeH1WithWrist.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "UnitreeH1.zip"),
            "339417cef5051a912693f3c64d29dddc",
        )
        prefix = "UnitreeH1"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class UnitreeH1Usd(EmbodiChainDataset):
    """Dataset class for the Unitree H1 robot USD version.

    Directory structure:
        UnitreeH1Usd/H1_usd
            h1.usd
            h1.usda

    Example usage:
        >>> from embodichain.data.robot_dataset import UnitreeH1Usd
        >>> dataset = UnitreeH1Usd()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("UnitreeH1Usd/H1_usd/h1.usd"))
        >>> print(get_data_path("UnitreeH1Usd/H1_usd/h1.usda"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "H1_usd.zip"),
            "9fc19f8c8b4a49398ec661e6ea9877ee",
        )
        prefix = "UnitreeH1Usd"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class ABB(EmbodiChainDataset):
    """Dataset class for the ABB robot.

    Reference:
        https://global.abb/

    Directory structure:
        ABB/
            IRB1200_5_90/IRB1200_5_90.urdf
            IRB2600_12_165/IRB2600_12_165.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import ABB
        >>> dataset = ABB()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("ABB/IRB1200_5_90/IRB1200_5_90.urdf"))
        >>> print(get_data_path("ABB/IRB2600_12_165/IRB2600_12_165.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "ABB.zip"),
            "ea6df4983982606c43387783e5fb8c05",
        )
        prefix = "ABB"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Motoman(EmbodiChainDataset):
    """Dataset class for the Motoman robot.

    Reference:
        https://www.motoman.com/en-us

    Directory structure:
        Motoman/
            GP7/GP7.urdf
            GP12/GP12.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Motoman
        >>> dataset = Motoman()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Motoman/GP7/GP7.urdf"))
        >>> print(get_data_path("Motoman/GP12/GP12.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Motoman.zip"),
            "ee5f16cfce34d8e2cb996fcff8a25986",
        )
        prefix = "Motoman"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class KUKA(EmbodiChainDataset):
    """Dataset class for the KUKA robot.

    Reference:
        https://www.kuka.com/

    Directory structure:
        KUKA/
            KUKA/KR6_R700_sixx/KR6_R700_sixx.urdf
            KUKA/KR6_R900_sixx/KR6_R900_sixx.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import ABB
        >>> dataset = ABB()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("KUKA/KR6_R700_sixx/KR6_R700_sixx.urdf"))
        >>> print(get_data_path("KUKA/KR6_R900_sixx/KR6_R900_sixx.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "KUKA.zip"),
            "da7a2dfd0db3f486e407f038d25c7537",
        )
        prefix = "KUKA"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Fanuc(EmbodiChainDataset):
    """Dataset class for the Fanuc robot.

    Reference:
        https://www.fanuc.com/

    Directory structure:
        Fanuc/
            M_20iA/M_20iA.urdf
            R_2000iC_165F/R_2000iC_165F.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Fanuc
        >>> dataset = Fanuc()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Fanuc/KR6_R700_sixx/KR6_R700_sixx.urdf"))
        >>> print(get_data_path("Fanuc/KR6_R900_sixx/KR6_R900_sixx.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Fanuc.zip"),
            "0a1c562f4719f7cdc1b24545fec4a301",
        )
        prefix = "Fanuc"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class UniversalRobots(EmbodiChainDataset):
    """Dataset class for the Universal Robots.

    Reference:
        https://www.universal-robots.com/products/ur-series/

    Directory structure:
        UniversalRobots/
            UR3/UR3.urdf
            UR3e/UR3e.urdf
            UR5/UR5.urdf
            UR5e/UR5e.urdf
            UR10/UR10.urdf
            UR10e/UR10e.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import UniversalRobots
        >>> dataset = UniversalRobots()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("UniversalRobots/UR3/UR3.urdf"))
        >>> print(get_data_path("UniversalRobots/UR3e/UR3e.urdf"))
        >>> print(get_data_path("UniversalRobots/UR5/UR5.urdf"))
        >>> print(get_data_path("UniversalRobots/UR5e/UR5e.urdf"))
        >>> print(get_data_path("UniversalRobots/UR10/UR10.urdf"))
        >>> print(get_data_path("UniversalRobots/UR10e/UR10e.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(
                EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "UniversalRobots_v2.zip"
            ),
            "4dd4ee5a86125364a9e8cbcd4aafc31a",
        )
        prefix = "UniversalRobots"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Rokae(EmbodiChainDataset):
    """Dataset class for the Rokae robots.

    Reference:
        https://www.rokae.com/en/product/show/349/SR-Cobots.html

    Directory structure:
        Rokae/
            SR3/SR3.urdf
            SR5/SR5.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Rokae
        >>> dataset = Rokae()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Rokae/SR3/SR3.urdf"))
        >>> print(get_data_path("Rokae/SR5/SR5.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Rokae.zip"),
            "fbfb852d6139e94b7c422771542f988f",
        )
        prefix = "Rokae"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Franka(EmbodiChainDataset):
    """Dataset class for the Franka robots.

    Reference:
        https://franka.de/franka-research-3

    Directory structure:
        Franka/
            Panda/
                Panda.urdf
                PandaHand.urdf
                PandaWithHand.urdf
            FR3/
                fr3.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Franka
        >>> dataset = Franka()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Franka/Panda/Panda.urdf"))
        >>> print(get_data_path("Franka/PandaHand/PandaHand.urdf"))
        >>> print(get_data_path("Franka/PandaWithHand/PandaWithHand.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "FrankaV2.zip"),
            "f0675b9da98126bc3d4e18c98ef5e06c",
        )
        prefix = "Franka"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Agile(EmbodiChainDataset):
    """Dataset class for the Agile robots.

    Reference:
        https://www.agile-robots.com/en/solutions/diana-7/

    Directory structure:
        Agile/
            Diana7/Diana7.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Agile
        >>> dataset = Agile()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Agile/Diana7/Diana7.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Agile.zip"),
            "fd47d7ab8a4d13960fd76e59544ba836",
        )
        prefix = "Agile"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Hans(EmbodiChainDataset):
    """Dataset class for the Hans robots.

    Reference:
        https://www.huayan-robotics.com/elfin

    Directory structure:
        Hans/
            E05/E05.urdf
            E10/E10.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Hans
        >>> dataset = Hans()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Hans/E05/E05.urdf"))
        >>> print(get_data_path("Hans/E10/E10.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Hans.zip"),
            "c867c406e3dffd6982fd0a15e7dc7e29",
        )
        prefix = "Hans"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class Aubo(EmbodiChainDataset):
    """Dataset class for the Aubo robots.

    Reference:
        https://www.aubo-robotics.cn/

    Directory structure:
        Aubo/
            i5/i5.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import Aubo
        >>> dataset = Aubo()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("Aubo/i5/i5.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "Aubo.zip"),
            "2574649cd199c11267cc0f4aeac65557",
        )
        prefix = "Aubo"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class RainbowY1(EmbodiChainDataset):
    """Dataset class for the Aubo robots.

    Reference:
        https://www.rainbow-robotics.com/en_rby1

    Directory structure:
        RainbowY1/
            RainbowY1.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import RainbowY1
        >>> dataset = RainbowY1()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("RainbowY1/RainbowY1.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "RainbowY1.zip"),
            "5979a3aaadb5de6488b13765d523564f",
        )
        prefix = "RainbowY1"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class CartPole(EmbodiChainDataset):
    """Dataset class for the CartPole.

    Directory structure:
        cart_pole/
            cart_pole.urdf
            cart.mtl
            cart.obj
            pole.mtl
            pole.obj
            slide_bar.mtl
            slide_bar.obj

    Example usage:
        >>> from embodichain.data.robot_dataset import CartPole
        >>> dataset = CartPole()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("CartPole/cart_pole.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "cart_pole.zip"),
            "9d185eb18b19f9c95153e01943c5b0a2",
        )
        prefix = "cart_pole"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class ARX5(EmbodiChainDataset):
    """Dataset class for the ARX5 robot.

    Reference:
        https://arx-x.com/

    Directory structure:
        ARX5/
            X5A.urdf

    Example usage:
        >>> from embodichain.data.robot_dataset import ARX5
        >>> dataset = ARX5()
        or
        >>> from embodichain.data import get_data_path
        >>> print(get_data_path("ARX5/X5A.urdf"))
    """

    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, robot_assets, "ARX5.zip"),
            "da207ba65e21577aa8e8f349af63e608",
        )
        prefix = "ARX5"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch


class QposSeedSampler:
    """
    Standard joint seed sampler for IK solving.

    Generates joint seed samples for each target pose in a batch, including the provided seed and random samples within joint limits.

    Args:
        num_samples (int): Number of samples per batch (including the seed).
        dof (int): Degrees of freedom.
        device (torch.device): Target device.
    """

    def __init__(self, num_samples: int, dof: int, device: torch.device):
        self.num_samples = num_samples
        self.dof = dof
        self.device = device

    def sample(
        self,
        qpos_seed: torch.Tensor,
        lower_limits: torch.Tensor,
        upper_limits: torch.Tensor,
        batch_size: int,
    ) -> torch.Tensor:
        """Generate joint seed samples for IK solving.

        Args:
            qpos_seed (torch.Tensor): (batch_size, dof) or (1, dof) initial seed.
            lower_limits (torch.Tensor): (dof,) lower joint limits.
            upper_limits (torch.Tensor): (dof,) upper joint limits.
            batch_size (int): Batch size.

        Returns:
            torch.Tensor: (batch_size * num_samples, dof) joint seeds.
        """
        joint_seeds_list = []
        for i in range(batch_size):
            current_seed = (
                qpos_seed[i].unsqueeze(0)
                if qpos_seed.shape[0] == batch_size
                else qpos_seed
            )
            if self.num_samples > 1:
                rand_part = lower_limits + (upper_limits - lower_limits) * torch.rand(
                    (self.num_samples - 1, self.dof), device=self.device
                )
            else:
                rand_part = torch.empty((0, self.dof), device=self.device)
            seeds = torch.cat([current_seed, rand_part], dim=0)
            joint_seeds_list.append(seeds)
        return torch.cat(joint_seeds_list, dim=0)

    def repeat_target_xpos(
        self, target_xpos: torch.Tensor, num_samples: int
    ) -> torch.Tensor:
        """Repeat each target pose num_samples times for batch processing.

        Args:
            target_xpos (torch.Tensor): (batch_size, 4, 4) or (batch_size, 3, 3) target poses.
            num_samples (int): Number of repeats per batch.

        Returns:
            torch.Tensor: (batch_size * num_samples, 4, 4) or (batch_size * num_samples, 3, 3)
        """
        repeated_list = [
            target_xpos[i].unsqueeze(0).repeat(num_samples, 1, 1)
            for i in range(target_xpos.shape[0])
        ]
        return torch.cat(repeated_list, dim=0)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------


import cv2
import pickle
import argparse
import time
import torch
import functools
import open3d as o3d
import numpy as np

from tqdm import tqdm
from PIL import Image
from functools import wraps
from typing import Dict, List, Tuple, Callable, Any

from embodichain.utils.string import callable_to_string


@functools.lru_cache(maxsize=None)  # memoization
def get_func_tag(tagName):
    return TagDecorator(tagName)


# https://stackoverflow.com/questions/41834530/how-to-make-python-decorators-work-like-a-tag-to-make-function-calls-by-tag
class TagDecorator(object):
    def __init__(self, tagName):
        self.functions = {}
        self.tagName = tagName

    def __str__(self):
        return "<TagDecorator {tagName}>".format(tagName=self.tagName)

    def __call__(self, f):
        class_name = f.__qualname__.split(".")[0]
        if class_name in self.functions.keys():
            self.functions[class_name].update({f.__name__: f})
        else:
            self.functions.update({class_name: {f.__name__: f}})
        return f


def set_attributes_for_class(self, params=None):
    if params:
        for k, v in params.items():
            if k != "self" and not k.startswith("_"):
                setattr(self, k, v)


def timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()  # 记录开始时间
        result = func(*args, **kwargs)  # 执行被装饰的函数
        end_time = time.time()  # 记录结束时间
        elapsed_time = end_time - start_time  # 计算耗时
        # log_warning(
        #     f"Function '{func.__name__}' executed in {elapsed_time:.4f} seconds"
        # )
        return result  # 返回被装饰函数的执行结果

    return wrapper


from embodichain.utils.logger import log_warning, log_error


def snake_to_camel(name):
    import re

    name = re.sub("_([a-zA-Z])", lambda m: (m.group(1).upper()), name)
    name = re.sub("-+", "_", name)
    return name


def convert_bytes(d):
    if isinstance(d, dict):
        return {convert_bytes(k): convert_bytes(v) for k, v in d.items()}
    if isinstance(d, list):
        return [convert_bytes(i) for i in d]
    if isinstance(d, bytes):
        return d.decode("UTF-8")
    return d


def pad_to_chunk(x: np.ndarray, chunk_size: int) -> np.ndarray:
    if x.shape[0] < chunk_size:

        if len(x.shape) <= 2:
            x = np.concatenate(
                [
                    x,
                    np.tile(
                        x[-1:],
                        (chunk_size - x.shape[0], 1),
                    ),
                ],
                axis=0,
            )
        elif len(x.shape) == 3 or len(x.shape) == 4:
            x = np.concatenate(
                [
                    x,
                    np.tile(
                        x[-1:],
                        (
                            (chunk_size - x.shape[0], 1, 1, 1)
                            if len(x[:1].shape) == 4
                            else (chunk_size - x.shape[0], 1, 1)
                        ),
                    ),
                ],
                axis=0,
            )
        else:
            raise ValueError("Unsupported shape {}.".format(x.shape))

    assert x.shape[0] == chunk_size, "shape {} vs chunk_size {}.".format(
        x.shape, chunk_size
    )
    return x


def dict2args(d: Dict) -> argparse.ArgumentParser:
    args = argparse.Namespace(**d)
    return args


def parser2dict(args) -> Dict:
    return vars(args)


def change_nested_dict(dict, keys, mode: str = "update", value=None):
    """
    Update or delete a nested dictionary at a specific key.

    Args:
        dict (dict): The dictionary to update.
        keys (tuple): Tuple of keys to the target value.
        mode (str): Whether to delete or remove the given key-value pair.
        value: The new value to set.

    Returns:
        dict: The updated dictionary.
    """
    if mode == "update":
        if value is None:
            log_error("The value to be updated is None, please check.")
        else:
            if len(keys) == 1:
                dict[keys[0]] = value
            else:
                change_nested_dict(dict[keys[0]], keys[1:], "update", value)
    elif mode == "delete":
        if value is not None:
            log_warning(
                f"Under mode 'delete' only the keys to be removed need to be provided. But got a not-None vlaue {value}."
            )
        if len(keys) == 1:
            del dict[keys[0]]
        else:
            change_nested_dict(dict[keys[0]], keys[1:], "delete")
    else:
        log_error(f"Mode '{mode}; is noet realized yet.")

    return dict


def set_texture_to_material(material, texture: np.ndarray, env, type: str = "color"):
    if type == "color":
        # TODO: Currently, create texture for base color map without alpha has error.
        # should be fixed in the future.
        if texture.shape[-1] == 3:
            texture = np.concatenate(
                [texture, np.ones_like(texture[..., :1]) * 255], axis=-1
            )

        color_texture = env.create_color_texture(texture, has_alpha=True)
        if color_texture:
            material.get_inst().set_base_color_map(color_texture)
    else:
        log_error(f"Unsupported texture type: {type}. Only 'color' is supported.")


def get_random_real_image(base_path: str, read: bool = True) -> np.ndarray:
    import os, random

    # 随机选择一个子文件夹
    subfolders = [f.path for f in os.scandir(base_path) if f.is_dir()]
    selected_subfolder = random.choice(subfolders)

    # 随机选择一个图片文件
    image_files = [
        f.path
        for f in os.scandir(selected_subfolder)
        if f.is_file() and f.path.endswith((".png", ".jpg", ".jpeg"))
    ]
    selected_image_file = random.choice(image_files)

    # 读取图片
    if read:
        real_image = cv2.imread(selected_image_file)
        return real_image
    else:
        return selected_image_file


def read_all_folder_images(base_path: str) -> List[np.ndarray]:
    """Read all images from all subfolders under the base path.

    Args:
        base_path (str): The base directory containing subfolders with images.

    Returns:
        List[np.ndarray]: A list of images read from the subfolders.
    """
    import os

    images = []
    # 遍历所有子文件夹
    # First, collect all image files
    image_files = []
    for subdir, _, files in os.walk(base_path):
        for file in files:
            if file.endswith((".png", ".jpg", ".jpeg")):
                image_files.append(os.path.join(subdir, file))

    # Then process with progress bar
    for image_path in tqdm(image_files, desc="Loading images"):
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        if image is not None:
            images.append(image)
    return images


def reset_all_seeds(seed: int = 0):
    import torch
    import random
    import open3d as o3d

    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    o3d.utility.random.seed(seed)


def do_process_decorator(
    pre_process: bool | None = True, post_process: bool | None = True
):
    """A decorator to decorate :meth:`inference`. Usage and example is comming soon.

    Args:
        pre_process (bool | None, optional): whether do pre-process. Defaults to True.
        post_process (bool | None, optional): whether do post-process. Defaults to True.
    """

    def inner_decorator(func: Callable):
        def main_wrapper(self, *args, **kwargs):
            if pre_process:
                input = getattr(self, "pre_process")(*args, **kwargs)
            if isinstance(input, dict):
                ret = func(self, input)
            else:
                ret = func(self, *input)
            if post_process:
                output = getattr(self, "post_process")(*ret)
            return output

        return main_wrapper

    return inner_decorator


def pad_img_list(img_list, max_len):
    while len(img_list) < max_len:
        img_list.append(None)


def get_right_name(name: str):
    return name + "_r"


def read_video(video_path: str):
    video = cv2.VideoCapture(video_path)
    total_frame_count = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    length = total_frame_count
    fps = video.get(cv2.CAP_PROP_FPS)
    return video, fps, length


def create_video_writer(
    video_path: str, resolution: Tuple[int, int], fps: int
) -> cv2.VideoWriter:
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # 用于mp4格式的生成
    video_vis = cv2.VideoWriter(
        video_path,
        fourcc,
        fps,
        (resolution[1], resolution[0]),
    )
    return video_vis


def update_array(
    mat: np.ndarray, vec: np.ndarray, first_is_latest: bool = True
) -> np.ndarray:
    if first_is_latest:
        mat[1:, :] = mat[:-1, :]
        mat[0, :] = vec
        return mat
    else:
        mat[:-1, :] = mat[1:, :]
        mat[-1, :] = vec
        return mat


def save_pkl(path: str, content):
    with open(path, "wb") as f:  # open a text file
        pickle.dump(content, f)  # serialize the list


def load_pkl(
    path: str,
):
    with open(path, "rb") as f:
        content = pickle.load(f)
    return content


def save_json(path: str, data):
    import json

    with open(path, "w") as f:
        json.dump(data, f, indent=4)


def load_json(path: str) -> Dict:
    import json

    with open(path) as f:
        config = json.load(f)
    return config


def load_txt(path: str) -> str:
    with open(path, "r") as f:
        contents = f.read().strip()
    return contents


def encode_image(image: np.ndarray, format: str = "png"):
    import base64

    image_encode = cv2.imencode(f".{format}", image)[1]
    base64_image = base64.b64encode(image_encode).decode("utf-8")
    return base64_image


def inv_transform(transform: np.ndarray) -> np.ndarray:
    """inverse transformation

    Args:
        transform (np.array): [np.array of size [4 x 4]]

    Returns:
        np.array: [np.array of size [4 x 4]]
    """
    r = transform[:3, :3]
    t = transform[:3, 3].T
    inv_r = r.T
    inv_t = -inv_r @ t
    inv_pose = np.eye(4, dtype=np.float32)
    inv_pose[:3, :3] = inv_r
    inv_pose[:3, 3] = inv_t
    return inv_pose


def scale_image(image, scale=0.5):
    import cv2

    h, w = image.shape[:2]
    if image.dtype == np.uint8:
        return cv2.resize(
            image,
            (
                int(w * scale),
                int(h * scale),
            ),
        )
    elif image.dtype == np.bool_:

        image = image.astype(np.uint8)
        image = cv2.resize(
            image,
            (
                int(w * scale),
                int(h * scale),
            ),
        )
        return image.astype(np.bool_)


def padding_by_longest_edge(img: np.ndarray) -> np.ndarray:
    w, h, c = img.shape[:3]
    e = np.maximum(w, h)
    ret = np.zeros((e, e, c)).astype(img.dtype)
    ret[:w, :h] = img
    return ret


def center_crop(img: np.ndarray, dim: Tuple[int, int]) -> np.ndarray:
    """Returns center cropped image
    Args:
    img: image to be center cropped
    dim: dimensions (width, height) to be cropped
    """
    width, height = img.shape[1], img.shape[0]

    # process crop width and height for max available dimension
    crop_width = dim[0] if dim[0] < img.shape[1] else img.shape[1]
    crop_height = dim[1] if dim[1] < img.shape[0] else img.shape[0]
    mid_x, mid_y = int(width / 2), int(height / 2)
    cw2, ch2 = int(crop_width / 2), int(crop_height / 2)
    crop_img = img[mid_y - ch2 : mid_y + ch2, mid_x - cw2 : mid_x + cw2]
    return crop_img


def postprocess_small_regions(
    masks: np.ndarray,
    min_area: int,
    max_area: int,
) -> List[int]:
    """Filter masks based on area constraints.

    Args:
        masks: Array of binary masks or list of masks.
        min_area: Minimum area threshold (exclusive - areas must be strictly greater).
        max_area: Maximum area threshold (inclusive - areas can equal this value).

    Returns:
        List of indices for masks that meet the area constraints (min_area < area <= max_area).
    """
    n = len(masks) if isinstance(masks, list) else masks.shape[0]
    # Use list comprehension for more efficient filtering
    # Logic: area > min_area and area <= max_area (original behavior preserved)
    return [
        i for i in range(n) if min_area < masks[i].astype(np.uint8).sum() <= max_area
    ]


def mask_to_box(mask: np.ndarray) -> np.ndarray:
    from torchvision.ops import masks_to_boxes
    import torch

    bbox = (
        masks_to_boxes(torch.from_numpy(mask).unsqueeze(0))
        .squeeze(0)
        .numpy()
        .astype(np.int16)
    )
    return bbox


def remove_overlap_mask(
    masks: List[np.ndarray], keep_inner_threshold: float = 0.5, eps: float = 1e-5
) -> List[int]:
    keep_ids = []

    # Pre-compute areas once for efficiency
    areas = np.array([mask.astype(np.uint8).sum() for mask in masks])

    for i, maskA in enumerate(masks):
        keep = True
        for j, maskB in enumerate(masks):
            if i == j:
                # 同一个mask，跳过
                continue
            if areas[i] > areas[j]:
                # 大的包裹mask不能被过滤
                continue

            # 计算交集
            intersection = (maskA * maskB).sum()
            # 计算maskA的覆盖比例
            overlap_ratio = intersection / (areas[i] + eps)
            # maskA被maskB覆盖的面积比例达到threshold，不保留
            if overlap_ratio >= keep_inner_threshold:
                keep = False
                break

        if keep:
            keep_ids.append(i)

    return keep_ids


def encode_image_from_path(image_path: str):
    import base64

    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def check_shared_memory_exists(name):
    from multiprocessing import shared_memory

    try:
        shm = shared_memory.SharedMemory(name=name)
        return True
    except FileNotFoundError:
        return False


def get_class_instance(module_name, class_name, *args, **kwargs):
    """Get an instance of a class from a module.

    Args:
        module_name (str): The name of the module to import.
        class_name (str): The name of the class to instantiate.

    Returns:
        object: An instance of the specified class.
    """
    import importlib

    # Import the module
    module = importlib.import_module(module_name)
    # Get the class from the module
    cls = getattr(module, class_name)
    return cls


def key_in_nested_dict(d: Dict, key: str) -> bool:
    """Check if a key exists in a nested dictionary.

    Args:
        d (Dict): A dictionary that may contain nested dictionaries.
        key (str): The key to search for in the dictionary.

    Returns:
        bool: True if the key exists in the dictionary or any of its nested dictionaries, False otherwise.
    """
    if key in d:
        return True
    for value in d.values():
        if isinstance(value, dict):  # Check if the value is a nested dictionary
            if key_in_nested_dict(
                value, key
            ):  # Recursively check the nested dictionary
                return True
    return False


def class_to_dict(obj: object) -> dict[str, Any]:
    """Convert an object into dictionary recursively.

    Note:
        Ignores all names starting with "__" (i.e. built-in methods).

    Args:
        obj: An instance of a class to convert.

    Raises:
        ValueError: When input argument is not an object.

    Returns:
        Converted dictionary mapping.
    """
    # check that input data is class instance
    if not hasattr(obj, "__class__"):
        raise ValueError(f"Expected a class instance. Received: {type(obj)}.")
    # convert object to dictionary
    if isinstance(obj, dict):
        obj_dict = obj
    elif isinstance(obj, torch.Tensor):
        # We have to treat torch tensors specially because `torch.tensor.__dict__` returns an empty
        # dict, which would mean that a torch.tensor would be stored as an empty dict. Instead we
        # want to store it directly as the tensor.
        return obj
    elif hasattr(obj, "__dict__"):
        obj_dict = obj.__dict__
    else:
        return obj

    # convert to dictionary
    data = dict()
    for key, value in obj_dict.items():
        # disregard builtin attributes
        if key.startswith("__"):
            continue
        # check if attribute is callable -- function
        if callable(value):
            data[key] = callable_to_string(value)
        # check if attribute is a dictionary
        elif hasattr(value, "__dict__") or isinstance(value, dict):
            data[key] = class_to_dict(value)
        # check if attribute is a list or tuple
        elif isinstance(value, (list, tuple)):
            data[key] = type(value)([class_to_dict(v) for v in value])
        else:
            data[key] = value
    return data


def get_mesh_md5(mesh: o3d.t.geometry.TriangleMesh) -> str:
    """get mesh md5 unique key

    Args:
        mesh (o3d.geometry.TriangleMesh): mesh

    Returns:
        str: mesh md5 value.
    """
    import hashlib

    vert = np.array(mesh.vertex.positions.numpy(), dtype=float)
    face = np.array(mesh.triangle.indices.numpy(), dtype=float)
    mix = np.vstack([vert, face])
    return hashlib.md5(np.array2string(mix).encode()).hexdigest()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import dexsim
import numpy as np

from dataclasses import dataclass
from functools import cached_property
from typing import List, Sequence, Dict, Union, Tuple, Optional

from dexsim.engine import Articulation as _Articulation
from dexsim.types import (
    ArticulationFlag,
    ArticulationGPUAPIWriteType,
    ArticulationGPUAPIReadType,
)
from dexsim.engine import CudaArray, PhysicsScene

from embodichain.lab.sim import VisualMaterialInst, VisualMaterial
from embodichain.lab.sim.cfg import ArticulationCfg, JointDrivePropertiesCfg
from embodichain.lab.sim.common import BatchEntity
from embodichain.utils.math import (
    matrix_from_quat,
    quat_from_matrix,
    convert_quat,
    matrix_from_euler,
)
from embodichain.lab.sim.utility.sim_utils import (
    get_dexsim_drive_type,
    set_dexsim_articulation_cfg,
    is_rt_enabled,
)
from embodichain.lab.sim.utility.solver_utils import (
    create_pk_chain,
    create_pk_serial_chain,
)
from embodichain.utils import logger


@dataclass
class ArticulationData:
    """GPU data manager for articulation."""

    def __init__(
        self, entities: List[_Articulation], ps: PhysicsScene, device: torch.device
    ) -> None:
        """Initialize the ArticulationData.

        Args:
            entities (List[_Articulation]): List of DexSim Articulation objects.
            ps (PhysicsScene): The physics scene.
            device (torch.device): The device to use for the articulation data.
        """
        self.entities = entities
        self.ps = ps
        self.num_instances = len(entities)
        self.device = device

        # get gpu indices for the entities.
        # only meaningful when using GPU physics.
        self.gpu_indices = (
            torch.as_tensor(
                [entity.get_gpu_index() for entity in self.entities],
                dtype=torch.int32,
                device=self.device,
            )
            if self.device.type == "cuda"
            else None
        )

        self.dof = self.entities[0].get_dof()
        self.num_links = self.entities[0].get_links_num()
        self.link_names = self.entities[0].get_link_names()

        self._root_pose = torch.zeros(
            (self.num_instances, 7), dtype=torch.float32, device=self.device
        )
        self._root_lin_vel = torch.zeros(
            (self.num_instances, 3), dtype=torch.float32, device=self.device
        )
        self._root_ang_vel = torch.zeros(
            (self.num_instances, 3), dtype=torch.float32, device=self.device
        )

        max_num_links = (
            self.ps.gpu_get_articulation_max_link_count()
            if self.device.type == "cuda"
            else self.num_links
        )
        self._body_link_pose = torch.zeros(
            (self.num_instances, max_num_links, 7),
            dtype=torch.float32,
            device=self.device,
        )
        self._body_link_vel = torch.zeros(
            (self.num_instances, max_num_links, 6),
            dtype=torch.float32,
            device=self.device,
        )

        self._body_link_lin_vel = torch.zeros(
            (self.num_instances, max_num_links, 3),
            dtype=torch.float32,
            device=self.device,
        )
        self._body_link_ang_vel = torch.zeros(
            (self.num_instances, max_num_links, 3),
            dtype=torch.float32,
            device=self.device,
        )

        max_dof = (
            self.ps.gpu_get_articulation_max_dof()
            if self.device.type == "cuda"
            else self.dof
        )
        self._target_qpos = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._qpos = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._target_qvel = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._qvel = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._qacc = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._qf = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )

    @property
    def root_pose(self) -> torch.Tensor:
        """Get the root pose of the articulation.

        Returns:
            torch.Tensor: The root pose of the articulation with shape of (num_instances, 7).
        """
        if self.device.type == "cpu":
            # Fetch pose from CPU entities
            root_pose = torch.as_tensor(
                np.array([entity.get_local_pose() for entity in self.entities]),
                dtype=torch.float32,
                device=self.device,
            )
            xyzs = root_pose[:, :3, 3]
            quats = quat_from_matrix(root_pose[:, :3, :3])
            return torch.cat((xyzs, quats), dim=-1)
        else:
            self.ps.gpu_fetch_root_data(
                data=self._root_pose,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.ROOT_GLOBAL_POSE,
            )
            self._root_pose[:, :4] = convert_quat(self._root_pose[:, :4], to="wxyz")
            return self._root_pose[:, [4, 5, 6, 0, 1, 2, 3]]

    @property
    def root_lin_vel(self) -> torch.Tensor:
        """Get the linear velocity of the root link of the articulation.

        Returns:
            torch.Tensor: The linear velocity of the root link with shape of (num_instances, 3).
        """
        if self.device.type == "cpu":
            # Fetch linear velocity from CPU entities
            return torch.as_tensor(
                np.array(
                    [entity.get_root_link_velocity()[:3] for entity in self.entities]
                ),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_root_data(
                data=self._root_lin_vel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.ROOT_LINEAR_VELOCITY,
            )
            return self._root_lin_vel.clone()

    @property
    def root_ang_vel(self) -> torch.Tensor:
        """Get the angular velocity of the root link of the articulation.

        Returns:
            torch.Tensor: The angular velocity of the root link with shape of (num_instances, 3).
        """
        if self.device.type == "cpu":
            # Fetch angular velocity from CPU entities
            return torch.as_tensor(
                np.array(
                    [entity.get_root_link_velocity()[3:] for entity in self.entities]
                ),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_root_data(
                data=self._root_ang_vel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.ROOT_ANGULAR_VELOCITY,
            )
        return self._root_ang_vel.clone()

    @property
    def root_vel(self) -> torch.Tensor:
        """Get the velocity of the root link of the articulation.

        Returns:
            torch.Tensor: The velocity of the root link, concatenating linear and angular velocities.
        """
        return torch.cat((self.root_lin_vel, self.root_ang_vel), dim=-1)

    @property
    def qpos(self) -> torch.Tensor:
        """Get the current positions (qpos) of the articulation.

        Returns:
            torch.Tensor: The current positions of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch qpos from CPU entities
            return torch.as_tensor(
                np.array(
                    [entity.get_current_qpos() for entity in self.entities],
                ),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._qpos,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_POSITION,
            )
            return self._qpos[:, : self.dof].clone()

    @property
    def target_qpos(self) -> torch.Tensor:
        """Get the target positions (target_qpos) of the articulation.

        Returns:
            torch.Tensor: The target positions of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch target_qpos from CPU entities
            return torch.as_tensor(
                np.array(
                    [
                        entity.get_current_qpos(is_target=True)
                        for entity in self.entities
                    ],
                ),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._target_qpos,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_TARGET_POSITION,
            )
            return self._target_qpos[:, : self.dof].clone()

    @property
    def qvel(self) -> torch.Tensor:
        """Get the current velocities (qvel) of the articulation.

        Returns:
            torch.Tensor: The current velocities of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch qvel from CPU entities
            return torch.as_tensor(
                np.array([entity.get_current_qvel() for entity in self.entities]),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._qvel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_VELOCITY,
            )
            return self._qvel[:, : self.dof].clone()

    @property
    def target_qvel(self) -> torch.Tensor:
        """Get the target velocities (target_qvel) of the articulation.
        Returns:
            torch.Tensor: The target velocities of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch target_qvel from CPU entities
            return torch.as_tensor(
                np.array(
                    [
                        entity.get_current_qvel(is_target=True)
                        for entity in self.entities
                    ],
                ),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._target_qvel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_TARGET_VELOCITY,
            )
            return self._target_qvel[:, : self.dof].clone()

    @property
    def qacc(self) -> torch.Tensor:
        """Get the current accelerations (qacc) of the articulation.

        Returns:
            torch.Tensor: The current accelerations of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch qacc from CPU entities
            return torch.as_tensor(
                np.array([entity.get_current_qacc() for entity in self.entities]),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._qacc,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_ACCELERATION,
            )
            return self._qacc[:, : self.dof].clone()

    @property
    def qf(self) -> torch.Tensor:
        """Get the current forces (qf) of the articulation.

        Returns:
            torch.Tensor: The current forces of the articulation with shape of (num_instances, dof).
        """
        if self.device.type == "cpu":
            # Fetch qf from CPU entities
            return torch.as_tensor(
                np.array([entity.get_current_qf() for entity in self.entities]),
                dtype=torch.float32,
                device=self.device,
            )
        else:
            self.ps.gpu_fetch_joint_data(
                data=self._qf,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.JOINT_FORCE,
            )
            return self._qf[:, : self.dof].clone()

    @property
    def body_link_pose(self) -> torch.Tensor:
        """Get the pose of all links in the articulation.

        Returns:
            torch.Tensor: The poses of the links in the articulation with shape (N, num_links, 7).
        """
        if self.device.type == "cpu":
            from embodichain.lab.sim.utility import get_dexsim_arenas

            arenas = get_dexsim_arenas()
            for j, entity in enumerate(self.entities):

                link_pose = np.zeros((self.num_links, 4, 4), dtype=np.float32)
                for i, link_name in enumerate(self.link_names):
                    pose = entity.get_link_pose(link_name)
                    arena_pose = arenas[j].get_root_node().get_local_pose()
                    pose[:2, 3] -= arena_pose[:2, 3]
                    link_pose[i] = pose

                link_pose = torch.from_numpy(link_pose)
                xyz = link_pose[:, :3, 3]
                quat = quat_from_matrix(link_pose[:, :3, :3])
                self._body_link_pose[j][: self.num_links, :] = torch.cat(
                    (xyz, quat), dim=-1
                )
            return self._body_link_pose[:, : self.num_links, :]
        else:
            self.ps.gpu_fetch_link_data(
                data=self._body_link_pose,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.LINK_GLOBAL_POSE,
            )
            quat = convert_quat(self._body_link_pose[..., :4], to="wxyz")
            return torch.cat((self._body_link_pose[..., 4:], quat), dim=-1)

    @property
    def body_link_vel(self) -> torch.Tensor:
        """Get the velocities of all links in the articulation.

        Returns:
            torch.Tensor: The poses of the links in the articulation with shape (N, num_links, 6).
        """
        if self.device.type == "cpu":
            for i, entity in enumerate(self.entities):
                self._body_link_vel[i][: self.num_links] = torch.from_numpy(
                    entity.get_link_general_velocities()
                )
            return self._body_link_vel[:, : self.num_links, :]
        else:
            self.ps.gpu_fetch_link_data(
                data=self._body_link_lin_vel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.LINK_LINEAR_VELOCITY,
            )
            self.ps.gpu_fetch_link_data(
                data=self._body_link_ang_vel,
                gpu_indices=self.gpu_indices,
                data_type=ArticulationGPUAPIReadType.LINK_ANGULAR_VELOCITY,
            )
            self._body_link_vel[..., :3] = self._body_link_lin_vel
            self._body_link_vel[..., 3:] = self._body_link_ang_vel
        return self._body_link_vel[:, : self.num_links, :]

    @property
    def joint_stiffness(self) -> torch.Tensor:
        """Get the joint stiffness of the articulation.

        Returns:
            torch.Tensor: The joint stiffness of the articulation with shape (N, dof).
        """
        return torch.as_tensor(
            np.array([entity.get_drive()[0] for entity in self.entities]),
            dtype=torch.float32,
            device=self.device,
        )

    @property
    def joint_damping(self) -> torch.Tensor:
        """Get the joint damping of the articulation.

        Returns:
            torch.Tensor: The joint damping of the articulation with shape (N, dof).
        """
        return torch.as_tensor(
            np.array([entity.get_drive()[1] for entity in self.entities]),
            dtype=torch.float32,
            device=self.device,
        )

    @property
    def joint_friction(self) -> torch.Tensor:
        """Get the joint friction of the articulation.

        Returns:
            torch.Tensor: The joint friction of the articulation with shape (N, dof).
        """
        return torch.as_tensor(
            np.array([entity.get_drive()[4] for entity in self.entities]),
            dtype=torch.float32,
            device=self.device,
        )

    @cached_property
    def qpos_limits(self) -> torch.Tensor:
        """Get the joint position limits of the articulation.

        Returns:
            torch.Tensor: The joint position limits of the articulation with shape (N, dof, 2).
        """
        return torch.as_tensor(
            np.array([entity.get_joint_limits() for entity in self.entities]),
            dtype=torch.float32,
            device=self.device,
        )

    @cached_property
    def qvel_limits(self) -> torch.Tensor:
        """Get the joint velocity limits of the articulation.

        Returns:
            torch.Tensor: The joint velocity limits of the articulation with shape (N, dof).
        """
        # TODO: get joint velocity limits always returns zero?
        return torch.as_tensor(
            np.array(
                [entity.get_drive()[3] for entity in self.entities],
            ),
            dtype=torch.float32,
            device=self.device,
        )

    @cached_property
    def qf_limits(self) -> torch.Tensor:
        """Get the joint effort limits of the articulation.

        Returns:
            torch.Tensor: The joint effort limits of the articulation with shape (N, dof).
        """
        return torch.as_tensor(
            np.array(
                [entity.get_drive()[2] for entity in self.entities],
            ),
            dtype=torch.float32,
            device=self.device,
        )

    @cached_property
    def link_vert_face(self) -> Dict[str, Tuple[torch.Tensor, torch.Tensor]]:
        """Get the vertices and faces of all links in the articulation.

        Returns:
            Dict[str, Tuple[torch.Tensor, torch.Tensor]]:
                - key (str): The name of the link.
                - vertices (torch.Tensor): The vertices of the specified link with shape (V, 3).
                - faces (torch.Tensor): The faces of the specified link with shape (F, 3).
        """
        link_vert_face = dict()
        for link_name in self.link_names:
            verts, faces = self.entities[0].get_link_vert_face(link_name)
            vertices_tensor = torch.as_tensor(
                verts, dtype=torch.float32, device=self.device
            )
            faces_tensor = torch.as_tensor(faces, dtype=torch.int32, device=self.device)
            link_vert_face[link_name] = (vertices_tensor, faces_tensor)
        return link_vert_face


class Articulation(BatchEntity):
    """Articulation represents a batch of articulations in the simulation.

    An articulation is a collection of rigid bodies connected by joints. The joints can be either
    fixed or actuated. The joints can be of different types, such as revolute or prismatic.

    For fixed-base articulation, it can be a robot arm, door, etc.
    For floating-base articulation, it can be a humanoid, drawer, etc.

    Args:
        cfg (ArticulationCfg): Configuration for the articulation.
        entities (List[_Articulation], optional): List of articulation entities.
        device (torch.device, optional): Device to use (CPU or CUDA).
    """

    def __init__(
        self,
        cfg: ArticulationCfg,
        entities: List[_Articulation] = None,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        # Initialize world and physics scene
        self._world = dexsim.default_world()
        self._ps = self._world.get_physics_scene()

        self.cfg = cfg
        self._entities = entities
        self.device = device

        # Store all indices for batch operations
        self._all_indices = torch.arange(len(entities), dtype=torch.int32)

        if device.type == "cuda":
            self._world.update(0.001)

        self._data = ArticulationData(entities=entities, ps=self._ps, device=device)

        self.cfg: ArticulationCfg
        if self.cfg.init_qpos is None:
            self.cfg.init_qpos = torch.zeros(self.dof, dtype=torch.float32)

        # Determine if we should use USD properties or cfg properties.
        if not self.cfg.use_usd_properties:
            # Set articulation configuration in DexSim
            set_dexsim_articulation_cfg(entities, self.cfg)

            num_entities = len(entities)
            dof = self._data.dof
            default_cfg = JointDrivePropertiesCfg()
            self.default_joint_damping = torch.full(
                (num_entities, dof),
                default_cfg.damping,
                dtype=torch.float32,
                device=device,
            )
            self.default_joint_stiffness = torch.full(
                (num_entities, dof),
                default_cfg.stiffness,
                dtype=torch.float32,
                device=device,
            )
            self.default_joint_max_effort = torch.full(
                (num_entities, dof),
                default_cfg.max_effort,
                dtype=torch.float32,
                device=device,
            )
            self.default_joint_max_velocity = torch.full(
                (num_entities, dof),
                default_cfg.max_velocity,
                dtype=torch.float32,
                device=device,
            )
            self.default_joint_friction = torch.full(
                (num_entities, dof),
                default_cfg.friction,
                dtype=torch.float32,
                device=device,
            )
            self._set_default_joint_drive()
        else:
            # Read current properties from USD-loaded entities
            self.default_joint_stiffness = self._data.joint_stiffness.clone()
            self.default_joint_damping = self._data.joint_damping.clone()
            self.default_joint_friction = self._data.joint_friction.clone()
            self.default_joint_max_effort = self._data.qf_limits.clone()
            self.default_joint_max_velocity = self._data.qvel_limits.clone()

            # Write the USD properties back to cfg
            usd_drive_pros = self.cfg.drive_pros
            usd_drive_pros.stiffness = (
                self.default_joint_stiffness[0].cpu().numpy().tolist()
            )
            usd_drive_pros.damping = (
                self.default_joint_damping[0].cpu().numpy().tolist()
            )
            usd_drive_pros.friction = (
                self.default_joint_friction[0].cpu().numpy().tolist()
            )
            usd_drive_pros.max_effort = (
                self.default_joint_max_effort[0].cpu().numpy().tolist()
            )
            usd_drive_pros.max_velocity = (
                self.default_joint_max_velocity[0].cpu().numpy().tolist()
            )

        self.pk_chain = None
        if self.cfg.build_pk_chain:
            self.pk_chain = create_pk_chain(
                urdf_path=self.cfg.fpath, device=self.device
            )

        # For rendering purposes, each articulation can have multiple material instances associated with its links.
        self._visual_material: List[Dict[str, VisualMaterialInst]] = [
            {} for _ in range(len(entities))
        ]
        self.is_shared_visual_material = False

        # Stores mimic information for joints.
        self._mimic_info = entities[0].get_mimic_info()

        self.active_joint_ids = [i for i in range(self.dof) if i not in self.mimic_ids]

        # TODO: very weird that we must call update here to make sure the GPU indices are valid.
        if device.type == "cuda":
            self._world.update(0.001)

        super().__init__(cfg, entities, device)

        # set default collision filter
        self._set_default_collision_filter()

        # flag for collision visible node existence
        self._has_collision_visible_node_dict = dict()
        for link_name in self.link_names:
            self._has_collision_visible_node_dict[link_name] = False

    def __str__(self) -> str:
        parent_str = super().__str__()
        return parent_str + f" | dof: {self.dof} | num_links: {self.num_links}"

    @cached_property
    def dof(self) -> int:
        """Get the degree of freedom of the articulation.

        Returns:
            int: The degree of freedom of the articulation.
        """
        return self._data.dof

    @cached_property
    def active_dof(self) -> int:
        """Get the number of active degrees of freedom of the articulation.

        Returns:
            int: The number of active degrees of freedom of the articulation.
        """
        return len(self.active_joint_ids)

    @cached_property
    def num_links(self) -> int:
        """Get the number of links in the articulation.

        Returns:
            int: The number of links in the articulation.
        """
        return self._data.num_links

    @cached_property
    def link_names(self) -> List[str]:
        """Get the names of the links in the articulation.

        Returns:
            List[str]: The names of the links in the articulation.
        """
        return self._data.link_names

    @cached_property
    def root_link_name(self) -> str:
        """Get the name of the root link of the articulation.

        Returns:
            str: The name of the root link.
        """
        return self.entities[0].get_root_link_name()

    @cached_property
    def joint_names(self) -> List[str]:
        """Get the names of the joints in the articulation.

        Returns:
            List[str]: The names of the actived joints in the articulation.
        """
        return self._entities[0].get_actived_joint_names()

    @cached_property
    def active_joint_names(self) -> List[str]:
        """Get the names of the active joints in the articulation.

        Returns:
            List[str]: The names of the active joints in the articulation.
        """
        return [self.joint_names[i] for i in self.active_joint_ids]

    @cached_property
    def all_joint_names(self) -> List[str]:
        """Get the names of the joints in the articulation.

        Returns:
            List[str]: The names of the joints in the articulation.
        """
        return self._entities[0].get_joint_names()

    @property
    def body_data(self) -> ArticulationData:
        """Get the rigid body data manager for this rigid object.

        Returns:
            RigidBodyData: The rigid body data manager.
        """
        return self._data

    @property
    def root_state(self) -> torch.Tensor:
        """Get the root state of the articulation.

        Returns:
            torch.Tensor: The root state of the articulation with shape (N, 13).
        """
        root_pose = self.body_data.root_pose
        root_lin_vel = self.body_data.root_lin_vel
        root_ang_vel = self.body_data.root_ang_vel
        return torch.cat((root_pose, root_lin_vel, root_ang_vel), dim=-1)

    @property
    def body_state(self) -> torch.Tensor:
        """Get the body state of the articulation.

        Returns:
            torch.Tensor: The body state of the articulation with shape (N, num_links, 13).
        """
        body_pose = self.body_data.body_link_pose
        body_vel = self.body_data.body_link_vel
        return torch.cat((body_pose, body_vel), dim=-1)

    @property
    def mimic_ids(self) -> List[int | None]:
        """Get the mimic joint ids for the articulation.

        Returns:
            List[int | None]: The mimic joint ids.
        """
        return self._mimic_info.mimic_id.tolist()

    @property
    def mimic_parents(self) -> List[int | None]:
        """Get the mimic joint parent ids for the articulation.

        Returns:
            List[int | None]: The mimic joint parent ids.
        """
        return self._mimic_info.mimic_parent.tolist()

    @property
    def mimic_multipliers(self) -> List[float]:
        """Get the mimic joint multipliers for the articulation.

        Returns:
            List[float]: The mimic joint multipliers.
        """
        return self._mimic_info.mimic_multiplier.tolist()

    @property
    def mimic_offsets(self) -> List[float]:
        """Get the mimic joint offsets for the articulation.

        Returns:
            List[float]: The mimic joint offsets.
        """
        return self._mimic_info.mimic_offset.tolist()

    def _set_default_collision_filter(self) -> None:
        collision_filter_data = torch.zeros(
            size=(self.num_instances, 4), dtype=torch.int32
        )
        for i in range(self.num_instances):
            collision_filter_data[i, 0] = i
            collision_filter_data[i, 1] = 1
        self.set_collision_filter(collision_filter_data)

    def set_collision_filter(
        self, filter_data: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """set collision filter data for the rigid object.

        Args:
            filter_data (torch.Tensor): [N, 4] of int.
                First element of each object is arena id.
                If 2nd element is 0, the object will collision with all other objects in world.
                3rd and 4th elements are not used currently.

            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used. Defaults to None.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(filter_data):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match pose length {len(filter_data)}."
            )

        filter_data_np = filter_data.cpu().numpy().astype(np.uint32)
        for i, env_idx in enumerate(local_env_ids):
            self._entities[env_idx].set_collision_filter_data(filter_data_np[i])

    def set_local_pose(
        self, pose: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set local pose of the articulation.

        Args:
            pose (torch.Tensor): The local pose of the articulation with shape (N, 7) or (N, 4, 4).
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(pose):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match pose length {len(pose)}."
            )

        if self.device.type == "cpu":
            pose = pose.cpu()
            if pose.dim() == 2 and pose.shape[1] == 7:
                pose_matrix = torch.eye(4).unsqueeze(0).repeat(pose.shape[0], 1, 1)
                pose_matrix[:, :3, 3] = pose[:, :3]
                pose_matrix[:, :3, :3] = matrix_from_quat(pose[:, 3:7])
                for i, env_idx in enumerate(local_env_ids):
                    self._entities[env_idx].set_local_pose(pose_matrix[i])
            elif pose.dim() == 3 and pose.shape[1:] == (4, 4):
                for i, env_idx in enumerate(local_env_ids):
                    self._entities[env_idx].set_local_pose(pose[i])
            else:
                logger.log_error(
                    f"Invalid pose shape {pose.shape}. Expected (N, 7) or (N, 4, 4)."
                )

            # TODO: in manual physics mode, the update should be explicitly called after
            # setting the pose to synchronize the state to renderer.
            self._world.update(0.001)

        else:
            if pose.dim() == 2 and pose.shape[1] == 7:
                xyz = pose[:, :3]
                quat = convert_quat(pose[:, 3:7], to="xyzw")
            elif pose.dim() == 3 and pose.shape[1:] == (4, 4):
                xyz = pose[:, :3, 3]
                quat = quat_from_matrix(pose[:, :3, :3])
                quat = convert_quat(quat, to="xyzw")
            else:
                logger.log_error(
                    f"Invalid pose shape {pose.shape}. Expected (N, 7) or (N, 4, 4)."
                )

            # we should keep `pose_` life cycle to the end of the function.
            pose_ = torch.cat((quat, xyz), dim=-1)
            indices = self.body_data.gpu_indices[local_env_ids]
            self._ps.gpu_apply_root_data(
                data=pose_,
                gpu_indices=indices,
                data_type=ArticulationGPUAPIWriteType.ROOT_GLOBAL_POSE,
            )
            self._ps.gpu_compute_articulation_kinematic(gpu_indices=indices)

            # TODO: To be removed when gpu articulation data sync is supported.
            if is_rt_enabled() is False:
                self.body_data.body_link_pose
                link_pose = self.body_data._body_link_pose[local_env_ids]
                self._world.sync_poses_gpu_to_cpu(
                    link_pose=CudaArray(link_pose),
                    articulation_gpu_indices=CudaArray(indices),
                )

    def get_local_pose(self, to_matrix=False) -> torch.Tensor:
        """Get local pose (root link pose) of the articulation.

        Args:
            to_matrix (bool, optional): If True, return the pose as a 4x4 matrix. If False, return as (x, y, z, qw, qx, qy, qz). Defaults to False.

        Returns:
            torch.Tensor: The local pose of the articulation with shape (N, 7) or (N, 4, 4) depending on `to_matrix`.
        """
        pose = self.body_data.root_pose
        if to_matrix:
            xyz = pose[:, :3]
            mat = matrix_from_quat(pose[:, 3:7])
            pose = (
                torch.eye(4, dtype=torch.float32, device=self.device)
                .unsqueeze(0)
                .repeat(pose.shape[0], 1, 1)
            )
            pose[:, :3, 3] = xyz
            pose[:, :3, :3] = mat
        return pose

    def get_link_vert_face(self, link_name: str) -> Tuple[torch.Tensor, torch.Tensor]:
        """Get the vertices and faces of a specific link in the articulation.

        Args:
            link_name (str): The name of the link.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]:
                - vertices (torch.Tensor): The vertices of the specified link with shape (V, 3).
                - faces (torch.Tensor): The faces of the specified link with shape (F, 3).
        """
        if link_name not in self.link_names:
            logger.log_error(
                f"Link name {link_name} not found in {self.__class__.__name__}. Available links: {self.link_names}"
            )

        verts, faces = self.body_data.link_vert_face[link_name]
        return verts, faces

    def get_link_pose(
        self, link_name: str, env_ids: Sequence[int] | None = None, to_matrix=False
    ) -> torch.Tensor:
        """Get the pose of a specific link in the articulation.

        Args:
            link_name (str): The name of the link.
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
            to_matrix (bool, optional): If True, return the pose as a 4x4 matrix. If False, return as (x, y, z, qw, qx, qy, qz). Defaults to False.

        Returns:
            torch.Tensor: The pose of the specified link with shape (N, 7) or (N, 4, 4) depending on `to_matrix`.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if link_name not in self.link_names:
            logger.log_error(
                f"Link name {link_name} not found in {self.__class__.__name__}. Available links: {self.link_names}"
            )

        link_idx = self.link_names.index(link_name)
        link_pose = self.body_data.body_link_pose[local_env_ids, link_idx, :]

        if to_matrix:
            xyz = link_pose[:, :3]
            mat = matrix_from_quat(link_pose[:, 3:7])
            link_pose = (
                torch.eye(4, dtype=torch.float32, device=self.device)
                .unsqueeze(0)
                .repeat(link_pose.shape[0], 1, 1)
            )
            link_pose[:, :3, 3] = xyz
            link_pose[:, :3, :3] = mat
        return link_pose

    def get_qpos(self, target: bool = False) -> torch.Tensor:
        """Get the current positions (qpos) or target positions (target_qpos) of the articulation.

        Args:
            target (bool): If True, gets target positions for simulation. If False, gets current positions.

        Returns:
            torch.Tensor: Joint positions with shape (N, dof), where N is the number of environments.
        """
        return self.body_data.qpos if not target else self.body_data.target_qpos

    def set_qpos(
        self,
        qpos: torch.Tensor,
        joint_ids: Sequence[int] | None = None,
        env_ids: Sequence[int] | None = None,
        target: bool = True,
    ) -> None:
        """Set the joint positions (qpos) or target positions for the articulation.

        Args:
            qpos (torch.Tensor): Joint positions with shape (N, dof), where N is the number of environments.
            joint_ids (Sequence[int] | None, optional): Joint indices to apply the positions. If None, applies to all joints.
            env_ids (Sequence[int] | None): Environment indices to apply the positions. Defaults to all environments.
            target (bool): If True, sets target positions for simulation. If False, updates current positions directly.

        Raises:
            ValueError: If the length of `env_ids` does not match the length of `qpos`.
        """
        # TODO: Refactor this part to use a more generic and extensible approach,
        # such as a class decorator that can automatically convert ndarray to torch.Tensor
        # and handle dimension padding for specified member functions.
        # This will make the codebase cleaner and reduce repetitive type checks/conversions.
        # (e.g., support specifying which methods should be decorated for auto-conversion.)
        if not isinstance(qpos, torch.Tensor):
            qpos = torch.as_tensor(qpos, dtype=torch.float32, device=self.device)

        if joint_ids is None:
            local_joint_ids = torch.arange(
                self.dof, device=self.device, dtype=torch.int32
            )
        elif not isinstance(joint_ids, torch.Tensor):
            local_joint_ids = torch.as_tensor(
                joint_ids, dtype=torch.int32, device=self.device
            )
        else:
            local_joint_ids = joint_ids

        local_env_ids = self._all_indices if env_ids is None else env_ids

        # Make sure qpos is 2D tensor
        if qpos.dim() == 1:
            qpos = qpos.unsqueeze(0)

        if len(local_env_ids) != len(qpos):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match qpos length {len(qpos)}. "
                f"env_ids: {local_env_ids}, qpos.shape: {qpos.shape}"
            )

        if self.device.type == "cpu":
            for i, env_idx in enumerate(local_env_ids):
                setter = (
                    self._entities[env_idx].set_current_qpos
                    if target
                    else self._entities[env_idx].set_qpos
                )
                setter(qpos[i].numpy(), local_joint_ids.numpy())
        else:
            limits = self.body_data.qpos_limits[0].T
            # clamp qpos to limits
            lower_limits = limits[0][local_joint_ids]
            upper_limits = limits[1][local_joint_ids]
            qpos = qpos.clamp(lower_limits, upper_limits)

            data_type = (
                ArticulationGPUAPIWriteType.JOINT_TARGET_POSITION
                if target
                else ArticulationGPUAPIWriteType.JOINT_POSITION
            )

            # Always fetch the latest data to avoid stale values
            if target:
                qpos_set = self.body_data._target_qpos
            else:
                qpos_set = self.body_data._qpos

            indices = self.body_data.gpu_indices[local_env_ids]
            qpos_set[local_env_ids[:, None], local_joint_ids] = qpos
            self._ps.gpu_apply_joint_data(
                data=qpos_set,
                gpu_indices=indices,
                data_type=data_type,
            )

    def get_qvel(self, target: bool = False) -> torch.Tensor:
        """Get the current velocities (qvel) or target velocities (target_qvel) of the articulation.

        Args:
            target (bool): If True, gets target velocities for simulation. If False, gets current velocities. The default is False.

        Returns:
            torch.Tensor: The current velocities of the articulation.
        """
        return self.body_data.qvel if not target else self.body_data.target_qvel

    def set_qvel(
        self,
        qvel: torch.Tensor,
        joint_ids: Sequence[int] | None = None,
        env_ids: Sequence[int] | None = None,
        target: bool = True,
    ) -> None:
        """Set the velocities (qvel) or target velocities of the articulation.

        Args:
            qvel (torch.Tensor): The velocities with shape (N, dof).
            joint_ids (Sequence[int] | None, optional): Joint indices to apply the velocities. If None, applies to all joints.
            env_ids (Sequence[int] | None, optional): Environment indices. Defaults to all indices.
            If True, sets target positions for simulation. If False, updates current positions directly.

        Raises:
            ValueError: If the length of `env_ids` does not match the length of `qvel`.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(qvel):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match qvel length {len(qvel)}."
            )

        if joint_ids is None:
            local_joint_ids = torch.arange(
                self.dof, device=self.device, dtype=torch.int32
            )
        elif not isinstance(joint_ids, torch.Tensor):
            local_joint_ids = torch.as_tensor(
                joint_ids, dtype=torch.int32, device=self.device
            )
        else:
            local_joint_ids = joint_ids

        if self.device.type == "cpu":
            for i, env_idx in enumerate(local_env_ids):
                setter = (
                    self._entities[env_idx].set_current_qvel
                    if target
                    else self._entities[env_idx].set_qvel
                )
                setter(qvel[i].numpy(), local_joint_ids)
        else:
            data_type = (
                ArticulationGPUAPIWriteType.JOINT_TARGET_VELOCITY
                if target
                else ArticulationGPUAPIWriteType.JOINT_VELOCITY
            )

            # Always fetch the latest data to avoid stale values
            if target:
                qvel_set = self.body_data._target_qvel
            else:
                qvel_set = self.body_data._qvel

            indices = self.body_data.gpu_indices[local_env_ids]
            qvel_set[local_env_ids[:, None], local_joint_ids] = qvel
            self._ps.gpu_apply_joint_data(
                data=qvel_set,
                gpu_indices=indices,
                data_type=data_type,
            )

    def set_qf(
        self,
        qf: torch.Tensor,
        joint_ids: Sequence[int] | None = None,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """Set the generalized efforts (qf) of the articulation.

        Args:
            qf (torch.Tensor): The generalized efforts with shape (N, dof).
            joint_ids (Sequence[int] | None, optional): Joint indices to apply the efforts. If None, applies to all joints.
            env_ids (Sequence[int] | None, optional): Environment indices. Defaults to all indices.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(qf):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match qf length {len(qf)}."
            )

        if self.device.type == "cpu":
            local_joint_ids = np.arange(self.dof) if joint_ids is None else joint_ids
            for i, env_idx in enumerate(local_env_ids):
                setter = self._entities[env_idx].set_current_qf
                setter(qf[i].numpy(), local_joint_ids)
        else:
            indices = self.body_data.gpu_indices[local_env_ids]
            if joint_ids is None:
                qf_set = self.body_data._qf[local_env_ids]
                qf_set[:, : self.dof] = qf
            else:
                self.body_data.qf
                qf_set = self.body_data._qf[local_env_ids]
                qf_set[:, joint_ids] = qf
            self._ps.gpu_apply_joint_data(
                data=qf_set,
                gpu_indices=indices,
                data_type=ArticulationGPUAPIWriteType.JOINT_FORCE,
            )

    def set_drive(
        self,
        stiffness: torch.Tensor | None = None,
        damping: torch.Tensor | None = None,
        max_effort: torch.Tensor | None = None,
        max_velocity: torch.Tensor | None = None,
        friction: torch.Tensor | None = None,
        drive_type: str = "force",
        joint_ids: Sequence[int] | None = None,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """Set the drive properties for the articulation.

        Args:
            stiffness (torch.Tensor): The stiffness of the joint drive with shape (len(env_ids), len(joint_ids)).
            damping (torch.Tensor): The damping of the joint drive with shape (len(env_ids), len(joint_ids)).
            max_effort (torch.Tensor): The maximum effort of the joint drive with shape (len(env_ids), len(joint_ids)).
            max_velocity (torch.Tensor): The maximum velocity of the joint drive with shape (len(env_ids), len(joint_ids)).
            friction (torch.Tensor): The joint friction coefficient with shape (len(env_ids), len(joint_ids)).
            drive_type (str, optional): The type of drive to apply. Defaults to "force".
            joint_ids (Sequence[int] | None, optional): The joint indices to apply the drive to. If None, applies to all joints. Defaults to None.
            env_ids (Sequence[int] | None, optional): The environment indices to apply the drive to. If None, applies to all environments. Defaults to None.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids
        local_joint_ids = np.arange(self.dof) if joint_ids is None else joint_ids

        for i, env_idx in enumerate(local_env_ids):
            drive_args = {
                "drive_type": get_dexsim_drive_type(drive_type),
                "joint_ids": local_joint_ids,
            }
            if stiffness is not None:
                drive_args["stiffness"] = stiffness[i].cpu().numpy()
            if damping is not None:
                drive_args["damping"] = damping[i].cpu().numpy()
            if max_effort is not None:
                drive_args["max_force"] = max_effort[i].cpu().numpy()
            if max_velocity is not None:
                drive_args["max_velocity"] = max_velocity[i].cpu().numpy()
            if friction is not None:
                drive_args["joint_friction"] = friction[i].cpu().numpy()
            self._entities[env_idx].set_drive(**drive_args)

    def get_user_ids(self, link_name: str | None = None) -> torch.Tensor:
        """Get the user ids of the articulation.

        Args:
            link_name: (str | None): The name of the link. If None, returns user ids for all links.

        Returns:
            torch.Tensor: The user ids of the articulation with shape (N, 1) for given link_name or (N, num_links) if link_name is None.
        """
        return torch.as_tensor(
            np.array(
                [entity.get_user_ids(link_name) for entity in self._entities],
            ),
            dtype=torch.int32,
            device=self.device,
        )

    def clear_dynamics(self, env_ids: Sequence[int] | None = None) -> None:
        """Clear the dynamics of the articulation.

        Args:
            env_ids (Sequence[int] | None): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids
        if self.device.type == "cpu":
            zero_joint_data = np.zeros((len(local_env_ids), self.dof), dtype=np.float32)
            for i, env_idx in enumerate(local_env_ids):
                self._entities[env_idx].set_qvel(zero_joint_data[i])
                self._entities[env_idx].set_current_qvel(zero_joint_data[i])
                self._entities[env_idx].set_current_qf(zero_joint_data[i])
        else:
            zeros = torch.zeros(
                (len(local_env_ids), self.dof), dtype=torch.float32, device=self.device
            )
            indices = self.body_data.gpu_indices[local_env_ids]
            self._ps.gpu_apply_joint_data(
                data=zeros,
                gpu_indices=indices,
                data_type=ArticulationGPUAPIWriteType.JOINT_VELOCITY,
            )
            self._ps.gpu_apply_joint_data(
                data=zeros,
                gpu_indices=indices,
                data_type=ArticulationGPUAPIWriteType.JOINT_TARGET_VELOCITY,
            )
            self._ps.gpu_apply_joint_data(
                data=zeros,
                gpu_indices=indices,
                data_type=ArticulationGPUAPIWriteType.JOINT_FORCE,
            )

    def reallocate_body_data(self) -> None:
        """Reallocate body data tensors to match the current articulation state in the GPU physics scene."""
        if self.device.type == "cpu":
            logger.log_warning(f"Reallocating body data on CPU is not supported.")
            return

        max_dof = self._ps.gpu_get_articulation_max_dof()
        max_num_links = self._ps.gpu_get_articulation_max_link_count()
        self._data._qpos = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._target_qpos = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._qvel = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._target_qvel = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._qacc = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._qf = torch.zeros(
            (self.num_instances, max_dof), dtype=torch.float32, device=self.device
        )
        self._data._body_link_pose = torch.zeros(
            (self.num_instances, max_num_links, 7),
            dtype=torch.float32,
            device=self.device,
        )
        self._data._body_link_vel = torch.zeros(
            (self.num_instances, max_num_links, 6),
            dtype=torch.float32,
            device=self.device,
        )

        self._data._body_link_lin_vel = torch.zeros(
            (self.num_instances, max_num_links, 3),
            dtype=torch.float32,
            device=self.device,
        )
        self._data._body_link_ang_vel = torch.zeros(
            (self.num_instances, max_num_links, 3),
            dtype=torch.float32,
            device=self.device,
        )
        self.reset()

    def reset(self, env_ids: Sequence[int] | None = None) -> None:
        local_env_ids = self._all_indices if env_ids is None else env_ids
        num_instances = len(local_env_ids)
        self.cfg: ArticulationCfg
        pos = torch.as_tensor(
            self.cfg.init_pos, dtype=torch.float32, device=self.device
        )
        rot = (
            torch.as_tensor(self.cfg.init_rot, dtype=torch.float32, device=self.device)
            * torch.pi
            / 180.0
        )
        pos = pos.unsqueeze(0).repeat(num_instances, 1)
        rot = rot.unsqueeze(0).repeat(num_instances, 1)
        mat = matrix_from_euler(rot, "XYZ")
        pose = (
            torch.eye(4, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .repeat(num_instances, 1, 1)
        )
        pose[:, :3, 3] = pos
        pose[:, :3, :3] = mat
        self.set_local_pose(pose, env_ids=local_env_ids)

        qpos = torch.as_tensor(
            self.cfg.init_qpos, dtype=torch.float32, device=self.device
        )
        qpos = qpos.unsqueeze(0).repeat(num_instances, 1)
        self.set_qpos(qpos, target=False, env_ids=local_env_ids)
        # Set drive target to hold position.
        self.set_qpos(qpos, target=True, env_ids=local_env_ids)

        self.clear_dynamics(env_ids=local_env_ids)

        if self.device.type == "cuda":
            self._ps.gpu_compute_articulation_kinematic(
                gpu_indices=self.body_data.gpu_indices[local_env_ids]
            )

            # TODO: To be removed when gpu articulation data sync is supported.
            if is_rt_enabled() is False:
                self.body_data.body_link_pose
                link_pose = self.body_data._body_link_pose[local_env_ids]
                indices = self.body_data.gpu_indices[local_env_ids]
                self._world.sync_poses_gpu_to_cpu(
                    link_pose=CudaArray(link_pose),
                    articulation_gpu_indices=CudaArray(indices),
                )
        else:
            self._world.update(0.001)

    def _set_default_joint_drive(self) -> None:
        """Set default joint drive parameters based on the configuration."""
        import numbers
        from embodichain.utils.string import resolve_matching_names_values

        drive_props = [
            ("damping", self.default_joint_damping),
            ("stiffness", self.default_joint_stiffness),
            ("max_effort", self.default_joint_max_effort),
            ("max_velocity", self.default_joint_max_velocity),
            ("friction", self.default_joint_friction),
        ]

        for prop_name, default_array in drive_props:
            value = getattr(self.cfg.drive_pros, prop_name, None)
            if value is None:
                continue
            if isinstance(value, numbers.Number):
                default_array[:] = value
            else:
                try:
                    indices, _, values = resolve_matching_names_values(
                        value, self.joint_names
                    )
                    default_array[:, indices] = torch.as_tensor(
                        values, dtype=torch.float32, device=self.device
                    )
                except Exception as e:
                    logger.log_error(f"Failed to set {prop_name}: {e}")

        drive_pros = self.cfg.drive_pros
        if isinstance(drive_pros, dict):
            drive_type = drive_pros.get("drive_type", "none")
        else:
            drive_type = getattr(drive_pros, "drive_type", "none")

        # Apply drive parameters to all articulations in the batch
        self.set_drive(
            stiffness=self.default_joint_stiffness,
            damping=self.default_joint_damping,
            max_effort=self.default_joint_max_effort,
            max_velocity=self.default_joint_max_velocity,
            friction=self.default_joint_friction,
            drive_type=drive_type,
        )

    def compute_fk(
        self,
        qpos: torch.Tensor | np.ndarray | None,
        link_names: str | list[str] | tuple[str] | None = None,
        end_link_name: str | None = None,
        root_link_name: str | None = None,
        to_dict: bool = False,
        **kwargs,
    ) -> Union[torch.Tensor, dict[str, "pk.Transform3d"]]:
        """Compute the forward kinematics (FK) for the given joint positions.

        Args:
            qpos (torch.Tensor): Joint positions. Shape can be (dof,) for a single configuration or
                                (batch_size, dof) for batched configurations.
            link_names (Union[str, list[str], tuple[str]], optional): Names of the links for which FK is computed.
                                                                    If None, all links are considered.
            end_link_name (str, optional): Name of the end link for which FK is computed. If None, all links are considered.
            root_link_name (str, optional): Name of the root link for which FK is computed. Defaults to None.
            to_dict (bool, optional): If True, returns the FK result as a dictionary of Transform3d objects. Defaults to False.
            **kwargs: Additional keyword arguments for customization.

        Raises:
            RuntimeError: If the pk_chain is not initialized.
            TypeError: If an invalid type is provided for `link_names`.
            ValueError: If the shape of the resulting matrices is unexpected.

        Returns:
            torch.Tensor: The homogeneous transformation matrix/matrices for the specified links.
                        Shape is (batch_size, 4, 4) for batched input or (4, 4) for single input.
                        If `to_dict` is True, returns a dictionary of Transform3d objects instead.
        """
        frame_indices = None
        if self.pk_chain is None:
            logger.log_error("pk_chain is not initialized for this articulation.")

        # Adapt link_names to work with get_frame_indices
        if link_names is not None:
            if isinstance(link_names, str):
                # Single link name
                frame_indices = self.pk_chain.get_frame_indices(link_names)
            elif isinstance(link_names, (list, tuple)):
                # Multiple link names
                frame_indices = self.pk_chain.get_frame_indices(*link_names)
            else:
                raise TypeError(
                    f"Invalid type for link_names: {type(link_names)}. Expected str, list, or tuple."
                )

        if end_link_name is None and root_link_name is None:
            result = self.pk_chain.forward_kinematics(
                th=qpos, frame_indices=frame_indices
            )
        else:
            pk_serial_chain = create_pk_serial_chain(
                chain=self.pk_chain,
                root_link_name=root_link_name,
                end_link_name=end_link_name,
            )
            result = pk_serial_chain.forward_kinematics(th=qpos, end_only=True)

        if to_dict:
            return result

        # Extract transformation matrices
        if isinstance(result, dict):
            if link_names:
                matrices = torch.stack(
                    [result[name].get_matrix() for name in link_names], dim=0
                )
            else:
                link_name = end_link_name if end_link_name else list(result.keys())[-1]
                matrices = result[link_name].get_matrix()
        elif isinstance(result, list):
            matrices = torch.stack(
                [xpos.get_matrix().squeeze() for xpos in result], dim=0
            )
        else:
            matrices = result.get_matrix()

        # Ensure batch format
        if matrices.dim() == 2:
            matrices = matrices.unsqueeze(0)

        # Create result tensor with proper homogeneous coordinates
        if matrices.dim() == 4:  # Multiple links
            num_links, batch_size, _, _ = matrices.shape
            result = (
                torch.eye(4, device=self.device)
                .expand(num_links, batch_size, 4, 4)
                .clone()
            )
            result[:, :, :3, :] = matrices[:, :, :3, :]
            result = result.permute(1, 0, 2, 3)  # (batch_size, num_links, 4, 4)
        elif matrices.dim() == 3:  # Single link
            batch_size, _, _ = matrices.shape
            result = torch.eye(4, device=self.device).expand(batch_size, 4, 4).clone()
            result[:, :3, :] = matrices[:, :3, :]
        else:
            raise ValueError(f"Unexpected matrices shape: {matrices.shape}")

        return result

    def compute_jacobian(
        self,
        qpos: torch.Tensor | np.ndarray | None,
        end_link_name: str = None,
        root_link_name: str = None,
        locations: torch.Tensor | np.ndarray | None = None,
        jac_type: str = "full",
    ) -> torch.Tensor:
        """Compute the Jacobian matrix for the given joint positions using the pk_serial_chain.

        Args:
            qpos (torch.Tensor): The joint positions. Shape can be (dof,) for a single configuration
                                 or (batch_size, dof) for batched configurations.
            end_link_name (str, optional): The name of the end link for which the Jacobian is computed.
                                           Defaults to the last link in the chain.
            root_link_name (str, optional): The name of the root link for which the Jacobian is computed.
                                            Defaults to the first link in the chain.
            locations (torch.Tensor | np.ndarray, optional): Offset points relative to the end-effector
                                                                   frame for which the Jacobian is computed.
                                                                   Shape can be (batch_size, 3) or (3,) for a single offset.
                                                                   Defaults to None (origin of the end-effector frame).
            jac_type (str, optional): Specifies the part of the Jacobian to return:
                                      - 'full': Returns the full Jacobian (6, dof) or (batch_size, 6, dof).
                                      - 'trans': Returns only the translational part (3, dof) or (batch_size, 3, dof).
                                      - 'rot': Returns only the rotational part (3, dof) or (batch_size, 3, dof).
                                      Defaults to 'full'.

        Raises:
            RuntimeError: If the pk_chain is not initialized.
            ValueError: If an invalid `jac_type` is provided.

        Returns:
            torch.Tensor: The Jacobian matrix. Shape depends on the input:
                          - For a single link: (6, dof) or (batch_size, 6, dof).
                          - For multiple links: (num_links, 6, dof) or (num_links, batch_size, 6, dof).
                          The shape also depends on the `jac_type` parameter.
        """
        if self.pk_chain is None:
            logger.log_error("pk_chain is not initialized for this articulation.")

        if qpos is None:
            qpos = torch.zeros(self.dof, device=self.device)

        # Ensure qpos is a tensor on the correct device
        qpos = torch.as_tensor(qpos, dtype=torch.float32, device=self.device)

        # Default root and end link names if not provided
        frame_names = self.pk_chain.get_frame_names()
        if root_link_name is None:
            root_link_name = frame_names[0]  # Default to the first frame
        if end_link_name is None:
            end_link_name = frame_names[-1]  # Default to the last frame

        # Create pk_serial_chain
        pk_serial_chain = create_pk_serial_chain(
            chain=self.pk_chain,
            root_link_name=root_link_name,
            end_link_name=end_link_name,
        )

        # Compute the Jacobian using the kinematics chain
        J = pk_serial_chain.jacobian(th=qpos, locations=locations)

        # Handle jac_type to return the desired part of the Jacobian
        if jac_type == "trans":
            return J[:, :3, :] if J.dim() == 3 else J[:3, :]
        elif jac_type == "rot":
            return J[:, 3:, :] if J.dim() == 3 else J[3:, :]
        elif jac_type == "full":
            return J
        else:
            raise ValueError(
                f"Invalid jac_type '{jac_type}'. Must be 'full', 'trans', or 'rot'."
            )

    def set_visual_material(
        self,
        mat: VisualMaterial,
        env_ids: Sequence[int] | None = None,
        link_names: List[str] | None = None,
        shared: bool = False,
    ) -> None:
        """Set visual material for the rigid object.

        Args:
            mat (VisualMaterial): The material to set.
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
            link_names (List[str] | None, optional): List of link names to apply the material to. If None, applies to all links.
            shared (bool, optional): Whether to share the material instance across links and environments. Defaults to False.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids
        link_names = self.link_names if link_names is None else link_names

        if shared:
            if len(local_env_ids) != self.num_instances:
                logger.log_error(f"Cannot share material instance for partial env_ids.")

            for link_name in link_names:
                mat_inst = mat.create_instance(f"{mat.uid}_{self.uid}_{link_name}")
                for i, env_idx in enumerate(local_env_ids):
                    self._entities[env_idx].set_material(link_name, mat_inst.mat)
                    self._visual_material[env_idx][link_name] = mat_inst
            self.is_shared_visual_material = True
        else:
            for i, env_idx in enumerate(local_env_ids):
                for link_name in link_names:
                    mat_inst = mat.create_instance(
                        f"{mat.uid}_{self.uid}_{link_name}_{env_idx}"
                    )
                    self._entities[env_idx].set_material(link_name, mat_inst.mat)
                    self._visual_material[env_idx][link_name] = mat_inst
            self.is_shared_visual_material = False

    def get_visual_material_inst(
        self,
        env_ids: Sequence[int] | None = None,
        link_names: List[str] | None = None,
    ) -> List[Dict[str, VisualMaterialInst]]:
        """Get visual material instances for the rigid object.

        Args:
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
            link_names (List[str] | None, optional): List of link names to filter materials. If None, returns materials for all links.
        Returns:
            List[Dict[str, VisualMaterialInst]]: A list where each element corresponds to an environment and contains a dictionary mapping link names to their VisualMaterialInst.
        """
        if env_ids is None and link_names is None:
            return self._visual_material

        local_env_ids = self._all_indices if env_ids is None else env_ids
        link_names = self.link_names if link_names is None else link_names

        result = []
        for i, env_idx in enumerate(local_env_ids):
            if link_names is None:
                result.append(self._visual_material[env_idx])
            else:
                mat_dict = {
                    link_name: self._visual_material[env_idx][link_name]
                    for link_name in link_names
                    if link_name in self._visual_material[env_idx]
                }
                result.append(mat_dict)
        return result

    def set_physical_visible(
        self,
        visible: bool = True,
        link_names: List[str] | None = None,
        rgba: Sequence[float] | None = None,
    ):
        """set collision

        Args:
            visible (bool, optional): is collision body visible. Defaults to True.
            link_names (List[str] | None, optional): links to set visibility. Defaults to None.
            rgba (Sequence[float] | None, optional): collision body visible rgba. It will be defined at the first time the function is called. Defaults to None.
        """
        rgba = rgba if rgba is not None else (0.8, 0.2, 0.2, 0.7)
        if len(rgba) != 4:
            logger.log_error(f"Invalid rgba {rgba}, should be a sequence of 4 floats.")
        rgba = np.array(
            [
                rgba[0],
                rgba[1],
                rgba[2],
                rgba[3],
            ]
        )
        link_names = self.link_names if link_names is None else link_names

        # create collision visible node if not exist
        if visible:
            for i, env_idx in enumerate(self._all_indices):
                for link_name in link_names:
                    if self._has_collision_visible_node_dict[link_name] is False:
                        self._entities[env_idx].create_physical_visible_node(
                            rgba, link_name
                        )
                        self._has_collision_visible_node_dict[link_name] = True

        # set visibility
        for i, env_idx in enumerate(self._all_indices):
            for link_name in link_names:
                self._entities[env_idx].set_physical_visible(visible, link_name)

    def set_fix_base(
        self,
        fix: bool = True,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """Set whether the base of the articulation is fixed.

        Args:
            fix (bool, optional): Whether to fix the base. Defaults to True.
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids
        for i, env_idx in enumerate(local_env_ids):
            self._entities[env_idx].set_articulation_flag(
                ArticulationFlag.FIX_BASE, fix
            )

    def set_self_collision(
        self,
        enable: bool = False,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """Set whether self-collision is enabled for the articulation.

        Args:
            enable (bool, optional): Whether to enable self-collision. Defaults to True.
            env_ids (Sequence[int] | None, optional): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids
        for i, env_idx in enumerate(local_env_ids):
            self._entities[env_idx].set_articulation_flag(
                ArticulationFlag.DISABLE_SELF_COLLISION, not enable
            )

    def destroy(self) -> None:
        env = self._world.get_env()
        arenas = env.get_all_arenas()
        if len(arenas) == 0:
            arenas = [env]
        for i, entity in enumerate(self._entities):
            arenas[i].remove_articulation(entity)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np

from embodichain.utils import logger, configclass
from embodichain.lab.sim.planners.utils import TrajectorySampleMethod
from embodichain.lab.sim.planners.base_planner import (
    validate_plan_options,
    BasePlanner,
    BasePlannerCfg,
    PlanOptions,
)
from .utils import PlanState, PlanResult

try:
    import toppra as ta
    import toppra.constraint as constraint
except ImportError:
    logger.log_error(
        "toppra not installed. Install with `pip install toppra==0.6.3`", ImportError
    )

ta.setup_logging(level="WARN")


__all__ = ["ToppraPlanner", "ToppraPlannerCfg", "ToppraPlanOptions"]


@configclass
class ToppraPlannerCfg(BasePlannerCfg):

    planner_type: str = "toppra"


@configclass
class ToppraPlanOptions(PlanOptions):

    constraints: dict = {
        "velocity": 0.2,
        "acceleration": 0.5,
    }
    """Constraints for the planner, including velocity and acceleration limits. 

    Should be a dictionary with keys 'velocity' and 'acceleration', each containing a value or a list of limits for each joint.
    """

    sample_method: TrajectorySampleMethod = TrajectorySampleMethod.QUANTITY
    """Method for sampling the trajectory. 
    
    Options are 'time' for uniform time intervals or 'quantity' for a fixed number of samples.
    """

    sample_interval: float | int = 0.01
    """Interval for sampling the trajectory. 
    
    If sample_method is 'time', this is the time interval in seconds. 
    If sample_method is 'quantity', this is the total number of samples.
    """


class ToppraPlanner(BasePlanner):
    def __init__(self, cfg: ToppraPlannerCfg):
        r"""Initialize the TOPPRA trajectory planner.

        References:
            - TOPPRA: Time-Optimal Path Parameterization for Robotic Systems (https://github.com/hungpham2511/toppra)

        Args:
            cfg: Configuration object containing ToppraPlanner settings
        """
        super().__init__(cfg)

        if self.robot.num_instances > 1:
            logger.log_error(
                "ToppraPlanner does not support multiple robot instances",
                NotImplementedError,
            )

    @validate_plan_options(options_cls=ToppraPlanOptions)
    def plan(
        self,
        target_states: list[PlanState],
        options: ToppraPlanOptions = ToppraPlanOptions(),
    ) -> PlanResult:
        r"""Execute trajectory planning.

        Args:
            target_states: List of dictionaries containing target states
            cfg: ToppraPlanOptions

        Returns:
            PlanResult containing the planned trajectory details.
        """

        for i, target in enumerate(target_states):
            if target.qpos is None:
                logger.log_error(f"Target state at index {i} missing qpos")

        dofs = len(target_states[0].qpos)

        # set constraints
        if isinstance(options.constraints["velocity"], float):
            vlims = np.array(
                [
                    [
                        -options.constraints["velocity"],
                        options.constraints["velocity"],
                    ]
                    for _ in range(dofs)
                ]
            )
        else:
            vlims = np.array(options.constraints["velocity"])

        if isinstance(options.constraints["acceleration"], float):
            alims = np.array(
                [
                    [
                        -options.constraints["acceleration"],
                        options.constraints["acceleration"],
                    ]
                    for _ in range(dofs)
                ]
            )
        else:
            alims = np.array(options.constraints["acceleration"])

        sample_method = options.sample_method
        sample_interval = options.sample_interval
        if sample_method == TrajectorySampleMethod.TIME and sample_interval <= 0:
            logger.log_error("Time interval must be positive", ValueError)
        if sample_method == TrajectorySampleMethod.TIME and isinstance(
            sample_interval, int
        ):
            logger.log_error(
                "Time interval must be a float when sample_method is TIME", TypeError
            )
        if sample_method == TrajectorySampleMethod.QUANTITY and sample_interval < 2:
            logger.log_error("At least 2 sample points required", ValueError)
        if sample_method == TrajectorySampleMethod.QUANTITY and isinstance(
            sample_interval, float
        ):
            logger.log_error(
                "Number of samples must be an integer when sample_method is QUANTITY",
                TypeError,
            )

        # Check waypoints
        for i, target in enumerate(target_states):
            if target.qpos is None:
                logger.log_error(f"Target state at index {i} missing qpos")
            if len(target.qpos) != dofs:
                logger.log_error(f"Target waypoints do not align at index {i}")

        if (
            len(target_states) == 2
            and torch.sum(torch.abs(target_states[1].qpos - target_states[0].qpos))
            < 1e-3
        ):
            logger.log_warning("Only two same waypoints, returning trivial trajectory.")
            return PlanResult(
                success=True,
                positions=torch.as_tensor(
                    np.stack([target_states[0].qpos, target_states[1].qpos]),
                    dtype=torch.float32,
                    device=self.device,
                ),
                velocities=torch.zeros(
                    (2, dofs), dtype=torch.float32, device=self.device
                ),
                accelerations=torch.zeros(
                    (2, dofs), dtype=torch.float32, device=self.device
                ),
                dt=torch.tensor([0.0, 0.0], dtype=torch.float32, device=self.device),
                duration=0.0,
            )

        # Build waypoints
        waypoints = []
        for target in target_states:
            waypoints.append(np.array(target.qpos))

        waypoints = np.array(waypoints)
        # Create spline interpolation
        # NOTE: Suitable for dense waypoints
        ss = np.linspace(0, 1, len(waypoints))

        # NOTE: Suitable for sparse waypoints; for dense waypoints, CubicSpline may fail strict monotonicity requirement
        # len_total = 0
        # len_from_start = [0]
        # for i in range(len(waypoints)-1):
        #     len_total += np.sum(np.abs(waypoints[i+1] - waypoints[i]))
        #     len_from_start.append(len_total)
        # ss = np.array([cur/len_total for cur in len_from_start])

        path = ta.SplineInterpolator(ss, waypoints)

        # Set constraints
        pc_vel = constraint.JointVelocityConstraint(vlims)
        pc_acc = constraint.JointAccelerationConstraint(alims)

        # Create TOPPRA instance
        instance = ta.algorithm.TOPPRA(
            [pc_vel, pc_acc],
            path,
            parametrizer="ParametrizeConstAccel",
            gridpt_min_nb_points=max(100, 10 * len(waypoints)),
        )
        # NOTES: Important to set a large number of grid points for better performance in dense waypoint scenarios.

        # Compute parameterized trajectory
        jnt_traj = instance.compute_trajectory()
        if jnt_traj is None:
            # raise RuntimeError("Unable to find feasible trajectory")
            logger.log_warning("Unable to find feasible trajectory")
            return PlanResult(success=False)

        duration = jnt_traj.duration
        # Sample trajectory points
        if duration <= 0:
            logger.log_error(f"Duration must be positive, got {duration}", ValueError)
        if sample_method == TrajectorySampleMethod.TIME:
            n_points = max(2, int(np.ceil(duration / sample_interval)) + 1)
            ts = np.linspace(0, duration, n_points)
        else:
            ts = np.linspace(0, duration, num=int(sample_interval))

        positions = []
        velocities = []
        accelerations = []

        for t in ts:
            positions.append(jnt_traj.eval(t))
            velocities.append(jnt_traj.evald(t))
            accelerations.append(jnt_traj.evaldd(t))

        dt = torch.as_tensor(ts, dtype=torch.float32, device=self.device)
        dt = torch.diff(dt, prepend=torch.tensor([0.0], device=self.device))

        return PlanResult(
            success=True,
            positions=torch.as_tensor(
                np.array(positions), dtype=torch.float32, device=self.device
            ),
            velocities=torch.as_tensor(
                np.array(velocities), dtype=torch.float32, device=self.device
            ),
            accelerations=torch.as_tensor(
                np.array(accelerations), dtype=torch.float32, device=self.device
            ),
            dt=dt,
            duration=duration,
        )



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np
from typing import List, Sequence
from dexsim.render import Light as _Light
from embodichain.lab.sim.cfg import LightCfg
from embodichain.lab.sim.common import BatchEntity
from embodichain.utils import logger


class Light(BatchEntity):
    """Light represents a batch of lights in the simulation.

    Each light supports the following properties:
        - Color (3 floats)
        - Intensity (1 float)
        - Falloff (1 float)
        - Location (3 floats)
    """

    def __init__(
        self,
        cfg: LightCfg,
        entities: List[_Light] = None,
        device: torch.device = torch.device("cpu"),
    ) -> None:

        super().__init__(cfg, entities, device)

    def set_color(
        self, colors: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set color for one or more lights.

        Args:
            colors (torch.Tensor): Tensor of shape (M, 3) or (3,), representing RGB values.
                - If shape is (3,), the same color is applied to all targeted instances.
                - If shape is (M, 3), M must match the number of targeted instances.
            env_ids (Sequence[int] | None): Indices of instances to set. If None:
                - For colors.shape == (3,), applies to all instances.
                - For colors.shape == (M, 3), M must equal num_instances, applies per-instance.
        """
        self._apply_vector3(colors, env_ids, "set_color")

    def set_intensity(
        self, intensities: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set intensity for one or more lights.

        Args:
            intensities (torch.Tensor): Tensor of shape (M,), (1,), or scalar (0-dim).
                - If scalar or shape (1,), the same intensity is applied to all targeted instances.
                - If shape (M,), M must match the number of targeted instances.
            env_ids (Sequence[int] | None): Indices of instances to set. If None:
                - For scalar/shape (1,), applies to all instances.
                - For shape (M,), M must equal num_instances, applies per-instance.
        """
        self._apply_scalar(intensities, env_ids, "set_intensity")

    def set_falloff(
        self, falloffs: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set falloff (radius) for one or more lights.

        Args:
            falloffs (torch.Tensor): Tensor of shape (M,), (1,), or scalar (0-dim).
                - If scalar or shape (1,), the same falloff is applied to all targeted instances.
                - If shape (M,), M must match the number of targeted instances.
            env_ids (Sequence[int] | None): Indices of instances to set. If None:
                - For scalar/shape (1,), applies to all instances.
                - For shape (M,), M must equal num_instances, applies per-instance.
        """
        self._apply_scalar(falloffs, env_ids, "set_falloff")

    def set_local_pose(
        self,
        pose: torch.Tensor,
        env_ids: Sequence[int] | None = None,
        to_matrix: bool = False,
    ) -> None:
        """Set local pose (translation) for one or more lights.

        Args:
            pose (torch.Tensor):
                - If to_matrix=False: shape (3,) or (M, 3), representing (x, y, z).
                - If to_matrix=True: shape (4, 4) or (M, 4, 4); translation extracted automatically.
            env_ids (Sequence[int] | None): Indices to set. If None:
                - For vector input (3,) broadcast to all, or (M,3) with M == num_instances.
                - For matrix input (4,4) broadcast to all, or (M,4,4) with M == num_instances.
            to_matrix (bool): Interpret `pose` as full 4x4 matrix if True, else as vector(s).
        """
        if not torch.is_tensor(pose):
            logger.log_error(
                f"set_local_pose requires a torch.Tensor, got {type(pose)}"
            )
            return

        cpu = pose.detach().cpu()
        if to_matrix:
            if cpu.ndim == 2 and cpu.shape == (4, 4):
                trans = cpu[:3, 3]
            elif cpu.ndim == 3 and cpu.shape[1:] == (4, 4):
                trans = cpu[..., 0:3, 3]
            else:
                logger.log_error(
                    f"set_local_pose matrix: expected (4,4) or (N,4,4), got {tuple(cpu.shape)}"
                )
                return
        else:
            trans = cpu  # expect (3,) or (M,3)

        try:
            self._apply_vector3(trans, env_ids, setter_name="set_location")
        except Exception as e:
            logger.log_error(f"set_local_pose: error while applying translation: {e}")

    def get_local_pose(self, to_matrix: bool = False) -> torch.Tensor:
        """Get local pose of each light, either as full matrix or translation vector.

        Args:
            to_matrix (bool, optional): If True, return poses as 4×4 matrices.
                If False, return translations only as (x, y, z). Defaults to False.
        Returns:
            torch.Tensor:
                - If to_matrix=True: Tensor of shape (N, 4, 4), where N == num_instances.
                - If to_matrix=False: Tensor of shape (N, 3), containing translations.
                On error or empty instances, returns an empty tensor with shape (0, 4, 4) or (0, 3) respectively, and logs via logger.log_error.
        """
        mats = []
        for i in range(self.num_instances):
            try:
                mat = self._entities[i].get_local_pose()  # expect numpy (4,4)
                arr = np.array(mat, dtype=np.float32)
                if arr.shape != (4, 4):
                    logger.log_error(
                        f"get_local_pose: unexpected shape {arr.shape} for instance {i}"
                    )
                    return torch.empty(
                        (0, 4, 4) if to_matrix else (0, 3), dtype=torch.float32
                    )
                mats.append(arr)
            except Exception as e:
                logger.log_error(f"get_local_pose: error for instance {i}: {e}")
                return torch.empty(
                    (0, 4, 4) if to_matrix else (0, 3), dtype=torch.float32
                )

        if not mats:
            return torch.empty((0, 4, 4) if to_matrix else (0, 3), dtype=torch.float32)

        stacked = np.stack(mats, axis=0)  # (N,4,4)
        tensor4 = torch.from_numpy(stacked)
        if to_matrix:
            return tensor4
        # else return translations
        return tensor4[:, 0:3, 3].clone()

    def _apply_vector3(
        self,
        tensor: torch.Tensor,
        env_ids: Sequence[int] | None,
        setter_name: str,
    ) -> None:
        """
        Generic helper for 3-element vectors (color, location).
        Expects tensor shape: (3,), or (M,3) with M == num_instances or M == len(env_ids).
        env_ids: Sequence[int] | None
        """
        # Validate tensor type
        if not torch.is_tensor(tensor):
            logger.log_error(
                f"{setter_name} requires a torch.Tensor, got {type(tensor)}"
            )
            return

        cpu = tensor.detach().cpu()
        # Determine target indices
        if env_ids is None:
            all_ids = list(range(self.num_instances))
        else:
            all_ids = list(env_ids)

        # Cases:
        # 1) cpu.ndim == 1 and size == 3: broadcast to all_ids
        if cpu.ndim == 1 and cpu.shape[0] == 3:
            arr = cpu.numpy()
            for i in all_ids:
                getattr(self._entities[i], setter_name)(
                    float(arr[0]), float(arr[1]), float(arr[2])
                )
            return

        # 2) cpu.ndim == 2 and cpu.shape == (num_instances, 3), env_ids None or full
        if cpu.ndim == 2 and cpu.shape == (self.num_instances, 3) and env_ids is None:
            arr_all = cpu.numpy()
            for i in range(self.num_instances):
                getattr(self._entities[i], setter_name)(
                    float(arr_all[i, 0]), float(arr_all[i, 1]), float(arr_all[i, 2])
                )
            return

        # 3) cpu.ndim == 2 and env_ids provided, cpu.shape == (len(env_ids), 3)
        if (
            cpu.ndim == 2
            and env_ids is not None
            and cpu.shape[0] == len(all_ids)
            and cpu.shape[1] == 3
        ):
            arr_sel = cpu.numpy()
            for idx, i in enumerate(all_ids):
                getattr(self._entities[i], setter_name)(
                    float(arr_sel[idx, 0]),
                    float(arr_sel[idx, 1]),
                    float(arr_sel[idx, 2]),
                )
            return

        logger.log_error(
            f"{setter_name}: tensor shape {tuple(cpu.shape)} is invalid for broadcasting "
            f"(expected (3,) or ({self.num_instances},3) or ({len(all_ids)},3))."
        )

    def _apply_scalar(
        self,
        tensor: torch.Tensor,
        env_ids: Sequence[int] | None,
        setter_name: str,
    ) -> None:
        """
        Generic helper for scalar floats (intensity, falloff).
        Accepts tensor shape: () (0-dim), (1,), or (M,) with M == num_instances or M == len(env_ids).
        env_ids: Sequence[int] | None
        """
        if not torch.is_tensor(tensor):
            logger.log_error(
                f"{setter_name} requires a torch.Tensor, got {type(tensor)}"
            )
            return

        cpu = tensor.detach().cpu()
        if env_ids is None:
            all_ids = list(range(self.num_instances))
        else:
            all_ids = list(env_ids)

        # 1) scalar tensor: broadcast
        if cpu.ndim == 0:
            val = float(cpu.item())
            for i in all_ids:
                getattr(self._entities[i], setter_name)(val)
            return

        # 2) 1D tensor:
        if cpu.ndim == 1:
            length = cpu.shape[0]
            arr = cpu.numpy()
            # a) length == num_instances and env_ids None: map one-to-one
            if length == self.num_instances and env_ids is None:
                for i in range(self.num_instances):
                    getattr(self._entities[i], setter_name)(float(arr[i]))
                return
            # b) length == len(env_ids) when env_ids provided: map one-to-one
            if env_ids is not None and length == len(all_ids):
                for idx, i in enumerate(all_ids):
                    getattr(self._entities[i], setter_name)(float(arr[idx]))
                return
            # c) length == 1: broadcast
            if length == 1:
                val = float(arr[0])
                for i in all_ids:
                    getattr(self._entities[i], setter_name)(val)
                return

        logger.log_error(
            f"{setter_name}: tensor shape {tuple(cpu.shape)} is invalid for broadcasting "
            f"(expected scalar, (1,), ({self.num_instances},) or ({len(all_ids)},))."
        )

    def reset(self, env_ids: Sequence[int] | None = None) -> None:
        self.cfg: LightCfg
        self.set_color(torch.as_tensor(self.cfg.color), env_ids=env_ids)
        self.set_intensity(torch.as_tensor(self.cfg.intensity), env_ids=env_ids)
        self.set_falloff(torch.as_tensor(self.cfg.radius), env_ids=env_ids)
        self.set_local_pose(torch.as_tensor(self.cfg.init_pos), env_ids=env_ids)
